package rossi.Component;

import java.awt.*;
import java.util.*;
import java.awt.image.*;

class TableCanvas extends Canvas
{
  private String[] ColumnHeaders;
  private Vector RowData;
  private int NumColumns;
  private int[] ColWidths;
  private Vector RowHeights;
  private Vector RowSelected;
  private Font f;
  private FontMetrics fm;
  private int TopRow=0, LeftCol=0;
  private int HeaderHeight;
  private int GrabbedColumnSeparator = -1;
  private int GrabbedRowSeparator = -1;
  private int SepXposition = 0;
  private int SepYposition = 0;
  private boolean RightMouseDown = false;
  private boolean LeftMouseDown = false;
  private boolean MultipleSelect = false;
  private int VisibleCols=0;
  private int VisibleRows=0;
  private int Highlighter = -1;
  private int StartDragRow, LastDragRow;

  TableCanvas()
  {
    Toolkit t;

    f = new Font("Courier", Font.PLAIN, 12);
    t = getToolkit();
    fm = t.getFontMetrics(f);
    RowHeights = new Vector();
    RowSelected = new Vector();
    RowData = new Vector();
    HeaderHeight = fm.getMaxAscent()+fm.getMaxDecent()+4;
    setBackground(Color.white);
  }

  public void setFont(Font fnt)
  {
    Toolkit t;

    f = fnt;
    t = getToolkit();
    fm = t.getFontMetrics(f);
    HeaderHeight = fm.getMaxAscent()+fm.getMaxDecent()+4;
  }

  public void setNumColumns(int n)
  {
    int i;

    NumColumns = n;
    
    ColumnHeaders = new String[n];
    ColWidths = new int[n];

    for (i=0; i<n; i++)
    {
      ColWidths[i] = 50;
      ColumnHeaders[i] = "";
    }
  }

  public void setColWidth(int Col, int Size)
  {
    ColWidths[Col] = Size; 
  }

  public void setHeaders(String[] Headers)
  {
    int i;

    for (i = 0; i<NumColumns; i++)
      ColumnHeaders[i] = Headers[i];
  }

  public void setTopRow(int r)
  {
    Graphics g;
    Rectangle rb = bounds();

    if (r < 0)
      return;

    if ((r - TopRow) == 1)
    {
      ScrollDownLine();
      TopRow = r;
      return;
    }

    if ((TopRow - r) == 1)
    {
      ScrollUpLine();
      TopRow = r;
      return;
    }

    TopRow = r;
    g = getGraphics();
    g.clipRect(0, HeaderHeight, rb.width, rb.height-HeaderHeight);
    g.clearRect(0, 0, rb.width, rb.height);
    paint(g);
  }

  public void setLeftCol(int c)
  {
    Graphics g;

    g = getGraphics();

    if ((c-LeftCol) == 1)  /* Scroll right */
    {
      Rectangle rb = bounds();
      g.copyArea(ColWidths[LeftCol], 0, rb.width-ColWidths[LeftCol], rb.height,
                 -ColWidths[LeftCol], 0);
      g.clipRect(rb.width-ColWidths[LeftCol], 0, ColWidths[LeftCol], rb.height);
      g.clearRect(rb.width-ColWidths[LeftCol], 0, ColWidths[LeftCol], rb.height);
    }

    LeftCol = c;
    paint(g);
  }

  public int getVisibleCols()
  {
    return VisibleCols;
  }

  public int getVisibleRows()
  {
    return VisibleRows;
  }

  private void drawCell(Graphics g, Color bgColor, Color fgColor, 
                        Object obj, int x, int y, int width, int height, 
                        boolean drawBorders)
  {
    g.setColor(bgColor);
    g.fillRect(x, y, width, height);
    g.setColor(fgColor);
    g.setFont(f);

    if (obj instanceof String)
    {
      g.drawString((String)(obj), x+2, y+2+fm.getMaxAscent());
    }

    if (obj instanceof Image)
    {
      g.drawImage((Image)(obj), x+2, y+2, this);
    }

    if (drawBorders)
    {
      g.drawLine(x+width-1,y,x+width-1,y+height-1);
      g.drawLine(x,y+height-1,x+width-1,y+height-1);
    }
  }

  private int drawRow(Graphics g, int y, int row)
  {
    Object obj;
    int RowHeight;
    Color bg, fg;
    int x, col;

    Rectangle r = bounds();

    RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
    for (x=0, col=LeftCol; col<NumColumns; col++)
    {
      obj = ((Vector)(RowData.elementAt(row))).elementAt(col);

      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
      {
	bg = Color.black;
	fg = Color.white;
      }
      else
      {
	bg = Color.white;
	fg = Color.black;
      }

      drawCell(g, bg, fg, obj, x, y,
	       ColWidths[col], 
	       RowHeight, true);
       
      x += ColWidths[col];
      if (x > r.width)
	break;
    }
    VisibleCols = col-LeftCol;
    y+=RowHeight;
    if (x < r.width)
    {
      g.drawLine(x,y-1,r.width,y-1);
    } 

    return y;
  }

  private int ComputeMinimumRowHeight(int row)
  {
    int h = 0;
    Vector v;
    Object obj;
    int i;

    v = (Vector) RowData.elementAt(row);
    h = fm.getMaxAscent()+fm.getMaxDecent()+4;
    for (i=0; i<v.size(); i++)
    {
      obj = v.elementAt(i);
      if (obj instanceof Image)
      {
        prepareImage((Image) obj, this);
        if ((checkImage((Image) obj, this) & ImageObserver.HEIGHT) == 0)
        {
          int j;
          for (j=0; j<600; j++)
          {
            if ((checkImage((Image) obj, this) & ImageObserver.HEIGHT) != 0)
              break;
            try {Thread.sleep(100);} catch (InterruptedException e) {}
          }
        }
        if ((((Image)obj).getHeight(this)+4) > h)
          h = ((Image)obj).getHeight(this) + 4;
      }
    }
    return h;
  }

  private synchronized void ComputeRowHeight(int row)
  {
    if (row >= RowHeights.size())
       RowHeights.addElement(new Object());

    RowHeights.setElementAt(new Integer(ComputeMinimumRowHeight(row)), row); 
  }

  public synchronized void addRow(Vector r)
  {
    RowData.addElement(r);
    RowSelected.addElement(new Boolean(false));
    ComputeRowHeight(RowData.size()-1);
  }

  public synchronized void addRow(Vector r, int row)
  {
    RowData.insertElementAt(r,row);
    RowSelected.insertElementAt(new Boolean(false),row);
    RowHeights.insertElementAt(new Object(), row);
    ComputeRowHeight(row);
    repaint();
  }

  public synchronized void deleteAll()
  {
    RowData.removeAllElements();
    RowSelected.removeAllElements();
    RowHeights.removeAllElements();
    TopRow = 0;
    LeftCol = 0;
    repaint();
  }

  public synchronized void deleteRow(int row)
  {
    RowData.removeElementAt(row);
    RowSelected.removeElementAt(row);
    RowHeights.removeElementAt(row);
    if (TopRow == row && TopRow != 0)
      TopRow--;
    repaint();
  }

  public int countRows()
  {
    return RowData.size();
  }

  public void selectRow(int rowsel, boolean select)
  {
    int i, sx, sy, row, RowHeight;
    Rectangle r = bounds();
    Graphics g;

    /* If not multi-select, deselect all invisible rows */
    if (MultipleSelect == false && select)
      for (row = 0; row < TopRow; row++)
	if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        {
          postEvent(new Event(this.getParent(), 
                        Event.LIST_DESELECT, new Integer(row)));
	  RowSelected.setElementAt(new Boolean(false), row);
        }
 
    /* Figure out what row we are on */ 
    
    g = getGraphics();
    sy = HeaderHeight;
    for (row=TopRow; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      sy += RowHeight;
      if (row == rowsel)
      {
        if (select == false)
        {
          postEvent(new Event(this.getParent(),
                        Event.LIST_DESELECT, new Integer(row)));
          RowSelected.setElementAt(new Boolean(false), row);
        }
        else
        {
          postEvent(new Event(this.getParent(), 
                        Event.LIST_SELECT, new Integer(row)));
          RowSelected.setElementAt(new Boolean(true), row);
        }
 
        drawRow(g, sy-RowHeight, row);
        break;
      }
      else
      {
        if (MultipleSelect == false)
	  if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	  {
            postEvent(new Event(this.getParent(), 
                          Event.LIST_DESELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(false), row);
            drawRow(g, sy-RowHeight, row);
	  }
      }
    }

    for (row++; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      sy += RowHeight;

      if (MultipleSelect == false)
	if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	{
          postEvent(new Event(this.getParent(), 
                        Event.LIST_DESELECT, new Integer(row)));
	  RowSelected.setElementAt(new Boolean(false), row);
          if ((sy-RowHeight) < r.height)
	    drawRow(g, sy-RowHeight, row);
        }
    }
  }

  public Vector getRow(int row)
  {
    return (Vector) RowData.elementAt(row);
  }

  public int getSelectedIndex()
  {
    int row;

    for (row = 0; row < RowData.size(); row++)
      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        return row;

    return -1;
  }

  public int[] getSelectedIndexes()
  {
    int row;
    int selected[];
    int numSelected = 0;
    int i;

    /* First count how many are selected */
    for (row = 0; row < RowData.size(); row++)
      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        numSelected++;

    selected = new int[numSelected];

    i = 0;
    for (row = 0; row < RowData.size(); row++)
      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        selected[i++] = row;

    return selected;
  }
 
  public Vector getSelectedRow()
  {
    int row;

    for (row = 0; row < RowData.size(); row++)
      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        return (Vector)RowData.elementAt(row);

    return null;
  }

  public Vector getSelectedRows()
  {
    Vector v;
    int row;

    v = new Vector();
    for (row = 0; row < RowData.size(); row++)
      if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        v.addElement(RowData.elementAt(row));

    return null;
  }

  public boolean isSelected(int row)
  {
    return ((Boolean)(RowSelected.elementAt(row))).booleanValue();
  }

  public void setMultipleSelections(boolean m)
  {
    MultipleSelect = m;
  }

  public synchronized void paint(Graphics g)
  {
    int col, row, x, y;

    Rectangle r = bounds();


    /* Draw the background for the headers */
    g.setColor(Color.blue);
    g.fillRect(0, 0, r.width, HeaderHeight);
    g.setColor(Color.black);


    /* Draw the headers */
    x = 0;
    for (col=LeftCol; col<NumColumns; col++)
    {
      drawCell(g, Color.blue, Color.white, ColumnHeaders[col], x, 0,
               ColWidths[col], 
               HeaderHeight, false);
      x += ColWidths[col];
      if (x > r.width)
        break;
    }

    /* Draw the data */
    for (y=HeaderHeight, row=TopRow; row<RowData.size(); row++)
    {
      y = drawRow(g, y, row);
      if (y > r.height)
        break;
    }

    VisibleRows = row - TopRow;

    if (Highlighter != -1)
      showHighlighter();
  }

  public boolean showHighlighter()
  {
    boolean highlighterVisible = false;
    int y, row, RowHeight=0;
    Graphics g;
    Rectangle r = bounds();

    /* Check to see if highligher is in visible range */
    if (Highlighter >= TopRow && Highlighter < (TopRow+VisibleRows))
      highlighterVisible = true;

    if (highlighterVisible == false)
      Highlighter = TopRow;

    g = getGraphics();
    g.setColor(Color.red); 

    /* Find the row the highlighter is on */
    for (y=HeaderHeight, row=TopRow; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      if (row == Highlighter)
        break;
      y += RowHeight;
    }

    g.drawLine(0, y-1, r.width, y-1);
    g.drawLine(0, y+RowHeight-1, r.width, y+RowHeight-1);
    g.setPaintMode();
    return highlighterVisible;
  }

  public int getHighlighted()
  {
    if (Highlighter >= TopRow && Highlighter < (TopRow+VisibleRows))
      return Highlighter;
    return -1;
  }

  public boolean highlighterDown()
  {
    int y, row, RowHeight=0;
    Graphics g;
    Rectangle r = bounds();

    /* Check to see if highligher is on last row */
    if (Highlighter == (RowData.size()-1))
      return true;

    /* Check to see if highligher is on last visible row */
    if (Highlighter == (TopRow+VisibleRows-1))
      return false;

    g = getGraphics();
    g.setColor(Color.black); 

    /* Find the row the highlighter is on */
    for (y=HeaderHeight, row=TopRow; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      if (row == Highlighter)
        break;
      y += RowHeight;
    }

    if (row != (RowData.size() - 1))
    {
      g.drawLine(0, y-1, r.width, y-1);
      g.drawLine(0, y+RowHeight-1, r.width, y+RowHeight-1);
      row++;
      y += RowHeight;
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      g.setColor(Color.red);
      g.drawLine(0, y-1, r.width, y-1);
      g.drawLine(0, y+RowHeight-1, r.width, y+RowHeight-1);
      Highlighter++;
    }
    g.setPaintMode();
    return true;
  }

  public boolean highlighterUp()
  {
    int y, row, RowHeight=0;
    Graphics g;
    Rectangle r = bounds();

    /* Check to see if highligher is on last visible row */
    if (Highlighter == TopRow)
      return false;

    g = getGraphics();
    g.setColor(Color.black); 

    /* Find the row the highlighter is on */
    for (y=HeaderHeight, row=TopRow; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      if (row == Highlighter)
        break;
      y += RowHeight;
    }

    if (row != 0)
    {
      g.drawLine(0, y-1, r.width, y-1);
      g.drawLine(0, y+RowHeight-1, r.width, y+RowHeight-1);
      row--;
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      y -= RowHeight;
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      g.setColor(Color.red);
      g.drawLine(0, y-1, r.width, y-1);
      g.drawLine(0, y+RowHeight-1, r.width, y+RowHeight-1);
      Highlighter--;
    }
    g.setPaintMode();
    return true;
  }

  public void selectHighlighted()
  {
    int i, sx, sy, row, RowHeight;
    Rectangle r = bounds();
    Graphics g;

    if (Highlighter == -1)
      return;

    /* If not multi-select, deselect all invisible rows */
    if (MultipleSelect == false)
      for (row = 0; row < TopRow; row++)
	if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        {
          postEvent(new Event(this.getParent(), 
                        Event.LIST_DESELECT, new Integer(row)));
	  RowSelected.setElementAt(new Boolean(false), row);
        }
 
    /* Figure out what row we are on */ 
    
    g = getGraphics();
    sy = HeaderHeight;
    for (row=TopRow; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      sy += RowHeight;
      if (row == Highlighter)
      {
        if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
        {
          postEvent(new Event(this.getParent(),
                        Event.LIST_DESELECT, new Integer(row)));
          RowSelected.setElementAt(new Boolean(false), row);
        }
        else
        {
          postEvent(new Event(this.getParent(), 
                        Event.LIST_SELECT, new Integer(row)));
          RowSelected.setElementAt(new Boolean(true), row);
        }
 
        drawRow(g, sy-RowHeight, row);
        break;
      }
      else
      {
        if (MultipleSelect == false)
	  if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	  {
            postEvent(new Event(this.getParent(), 
                          Event.LIST_DESELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(false), row);
            drawRow(g, sy-RowHeight, row);
	  }
      }
    }

    for (row++; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      sy += RowHeight;

      if (MultipleSelect == false)
	if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	{
          postEvent(new Event(this.getParent(), 
                        Event.LIST_DESELECT, new Integer(row)));
	  RowSelected.setElementAt(new Boolean(false), row);
          if ((sy-RowHeight) < r.height)
	    drawRow(g, sy-RowHeight, row);
        }
    }
  }

  private void ScrollUpLine()
  {
    int RowHeight, y, row;
    Graphics g;
    Rectangle r = bounds();
  
    if (TopRow == 0)
      return;

    g = getGraphics();

    if (TopRow < RowData.size())
      RowHeight = ((Integer)(RowHeights.elementAt(TopRow-1))).intValue();
    else
      RowHeight = 0;

    /* Blt down one line */
    g.copyArea(0, HeaderHeight, 
	       r.width, r.height-HeaderHeight, 
	       0, RowHeight);
    g.clearRect(0, HeaderHeight, r.width, RowHeight);

    /* Draw the new top row */
    drawRow(g, HeaderHeight, TopRow-1);

    /* Count the number of visible rows */

    for (y=HeaderHeight, row=TopRow-1; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      y += RowHeight;
      if (y > r.height)
        break;
    }

    VisibleRows = row-TopRow+1;
  }

  private void ScrollDownLine()
  {
    int y, row;
    int RowHeight;
    Graphics g;
    Rectangle r = bounds();
    int BottomRow, BottomRowY=0;


    /* Locate the last visible row */

    BottomRow = RowData.size()-1;
    for (y=HeaderHeight, row=TopRow+0; row<RowData.size(); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      y += RowHeight;
      BottomRowY = y - RowHeight;
      if (y > r.height)
      {
        BottomRow = row;
        break;
      }
    }
 
    /* If less than 3 rows visible, just paint */
    g = getGraphics();
    if ((BottomRow - TopRow - 1) < 3)
    {
      g.setColor(Color.white);
      g.fillRect(0, HeaderHeight, r.width, r.height - HeaderHeight);
      TopRow++;
      paint(g);
      TopRow--;
      return;
    }
 
    /* Blt the screen up one row */
    RowHeight = ((Integer)(RowHeights.elementAt(TopRow))).intValue();

    g.copyArea(0, HeaderHeight+RowHeight, 
	       r.width, BottomRowY-HeaderHeight-RowHeight, 
	       0, -RowHeight);

    g.clearRect(0, BottomRowY-RowHeight, r.width, r.height-BottomRowY+RowHeight);
    
    for (y=HeaderHeight, row=TopRow+1; row<(BottomRow-1); row++)
    {
      RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
      y += RowHeight;
    }
  
    for (row=BottomRow-1; row<RowData.size(); row++)
    {
      y = drawRow(g, y, row);
      if (y > r.height)
       break;
    }
 
//    VisibleRows = row - TopRow - 1;
  
    if (y < r.height)
    { 
      g.clearRect(0, y, r.width, r.height - y);
    }
  }

  /* Mouse is down */
  public boolean mouseDown(Event e, int x, int y)
  {
    int i, sx, sy;
    Rectangle r = bounds();

    /* Get keyboard input focus */
    requestFocus();

    if (e.metaDown() == false) /* Left mouse */
    {
      LeftMouseDown = true;
      StartDragRow = -1;
      LastDragRow = -1;

      /* Check for double click */
      if (e.clickCount == 2)
      {
	/* See what row we are on */
	for (i=TopRow, sy=HeaderHeight; i<RowData.size(); i++)
        {
	  sy += ((Integer)(RowHeights.elementAt(i))).intValue();
          if (sy > y)
            break;
        }

	postEvent(new Event(this.getParent(), 
		      Event.ACTION_EVENT, new Integer(i)));

        LeftMouseDown = false;
        return true;
      }

      /* See if we are on a column separator */
      for (i=LeftCol, sx=0; i<NumColumns; i++)
      {
	sx += ColWidths[i];
	if (x > (sx-2)  && x < (sx+2))
	{
          GrabbedRowSeparator = -1;
	  GrabbedColumnSeparator = i;
	  SepXposition = sx;
	  return true;
	}
	if (sx > r.width)
	  break; 
      }
      GrabbedColumnSeparator = -1;

      /* See if we are on a row separator */
      for (i=TopRow, sy=HeaderHeight; i<RowData.size(); i++)
      {
	sy += ((Integer)(RowHeights.elementAt(i))).intValue();
	if (y > (sy-2)  && y < (sy+2))
	{
          GrabbedColumnSeparator = -1;
	  GrabbedRowSeparator = i;
	  SepYposition = sy;
	  return true;
	}
	if (sy > r.height)
	  break; 
      }
      GrabbedRowSeparator = -1;
    }
    else
    { 
      RightMouseDown = true;
    }
    
    return true;
  }

  /* Mouse is up */
  public boolean mouseUp(Event e, int x, int y)
  {
    int i, sx, sy, row, RowHeight, MinRowHeight;
    Rectangle r = bounds();
    Graphics g;

    if (LeftMouseDown && !RightMouseDown) /* Left mouse */
    {
      if (GrabbedColumnSeparator > -1)
      {
	/* Figure out where we were */
	for (i=LeftCol, sx=0; i<=GrabbedColumnSeparator; i++)
	  sx += ColWidths[i];

	{
	  ColWidths[GrabbedColumnSeparator] += (x-sx);
	  if (ColWidths[GrabbedColumnSeparator] < 10)
	    ColWidths[GrabbedColumnSeparator] = 10;
	}
   
	repaint();

	return true;
      }

      if (GrabbedRowSeparator > -1)
      {
        MinRowHeight = ComputeMinimumRowHeight(GrabbedRowSeparator);
	/* Figure out where we were */
	for (i=TopRow, sy=HeaderHeight; i<=GrabbedRowSeparator; i++)
	  sy += ((Integer)(RowHeights.elementAt(i))).intValue();

	{
          RowHeight = ((Integer)(RowHeights.elementAt(GrabbedRowSeparator))).intValue();
          RowHeight += (y-sy);
	  if (RowHeight < MinRowHeight)
	    RowHeight = MinRowHeight;

          RowHeights.setElementAt(new Integer(RowHeight), GrabbedRowSeparator);
	}
   
	repaint();

	return true;
      }

      /* If not multi-select, deselect all invisible rows */
      if (MultipleSelect == false)
	for (row = 0; row < TopRow; row++)
	  if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
          {
	    postEvent(new Event(this.getParent(), 
			  Event.LIST_DESELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(false), row);
          }
   
      /* Figure out what row we are on */ 
      
      g = getGraphics();
      sy = HeaderHeight;
      for (row=TopRow; row<RowData.size(); row++)
      {
	RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
	sy += RowHeight;
	if (y < sy)
	{
          if (row == LastDragRow && row != StartDragRow)
            break;
	  if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
          {
	    postEvent(new Event(this.getParent(), 
			  Event.LIST_DESELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(false), row);
          }
	  else
          {
	    postEvent(new Event(this.getParent(), 
			  Event.LIST_SELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(true), row);
          }
   
	  drawRow(g, sy-RowHeight, row);
	  break;
	}
	else
	{
	  if (MultipleSelect == false)
	    if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	    {
	      postEvent(new Event(this.getParent(), 
			    Event.LIST_DESELECT, new Integer(row)));
	      RowSelected.setElementAt(new Boolean(false), row);
	      drawRow(g, sy-RowHeight, row);
	    }
	}
      }

      for (row++; row<RowData.size(); row++)
      {
	RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
	sy += RowHeight;

	if (MultipleSelect == false)
	  if (((Boolean)(RowSelected.elementAt(row))).booleanValue())
	  {
	    postEvent(new Event(this.getParent(), 
			  Event.LIST_DESELECT, new Integer(row)));
	    RowSelected.setElementAt(new Boolean(false), row);
	    if ((sy-RowHeight) < r.height)
	      drawRow(g, sy-RowHeight, row);
	  }
      }
      LeftMouseDown = false;
      return true;
    }

    LeftMouseDown = false;
    if (e.metaDown() == false)
      RightMouseDown = false;

    return false;
  }

  public boolean mouseDrag(Event e, int x, int y)
  {
    Graphics g;
    Rectangle r = bounds();
    
    if (GrabbedColumnSeparator > -1)
    {
      g = getGraphics();
      g.setXORMode(Color.yellow);
      g.drawLine(SepXposition, HeaderHeight, SepXposition, r.height);
      SepXposition = x;
      g.drawLine(SepXposition, HeaderHeight, SepXposition, r.height);
      g.setPaintMode();
    }

    if (GrabbedRowSeparator > -1)
    {
      g = getGraphics();
      g.setXORMode(Color.yellow);
      g.drawLine(0, SepYposition, r.width, SepYposition);
      SepYposition = y;
      g.drawLine(0, SepYposition, r.width, SepYposition);
      g.setPaintMode();
    }

    if (GrabbedRowSeparator == -1 && GrabbedColumnSeparator == -1)
    {
      int sy, row;
      int RowHeight;
      boolean select;

      sy = HeaderHeight;
      for (row=TopRow; row<RowData.size(); row++)
      {
	RowHeight = ((Integer)(RowHeights.elementAt(row))).intValue();
	sy += RowHeight;
        if (sy > r.height)
          break;

	if (y < sy)
        {
	  if (StartDragRow == -1)
	    StartDragRow = row;

          if (row == LastDragRow)
            break;

          select = true;
          if (row >= StartDragRow && row < LastDragRow)
          {
            row++;
            select = false;
          }

          if (row <= StartDragRow && row > LastDragRow && LastDragRow != -1)
          {
            row --;
            select = false;
          }

          selectRow(row, select);
          LastDragRow = row;
          break;
        }
      }
    }

    return true;
  }
}


/** Class that implements a Table.  This is like a List, but with
    multiple columns.  It can display a String or an Image in each
    of its cells. */

public class Table extends Panel
{
  private Scrollbar Horiz, Vert;
  private TableCanvas Selection;

  /** Creates a new Table */
  public Table()
  {
    Horiz = new Scrollbar(Scrollbar.HORIZONTAL); 
    Vert = new Scrollbar(Scrollbar.VERTICAL); 

    Vert.setLineIncrement(1);
    Vert.setPageIncrement(1);
    Vert.setValue(0);

    Horiz.setLineIncrement(1);
    Horiz.setPageIncrement(1);
    Horiz.setValue(0);

    Selection = new TableCanvas();

    setLayout(new BorderLayout(5,5));
    add("East", Vert);
    add("South", Horiz);    
    add("Center", Selection);
  }

  public boolean handleEvent(Event e)
  {

    if (e.id == Event.SCROLL_ABSOLUTE ||
        e.id == Event.SCROLL_LINE_DOWN ||
        e.id == Event.SCROLL_LINE_UP ||
        e.id == Event.SCROLL_PAGE_DOWN ||
        e.id == Event.SCROLL_PAGE_UP)
    {

      AdjustScrollbars();

      if (e.target == Horiz)
      {
	Selection.setLeftCol(Horiz.getValue());
	return true;
      }

      if (e.target == Vert)
      {
        if (Vert.getValue() != Vert.getMaximum())
	  Selection.setTopRow(Vert.getValue());
        else
	  Selection.setTopRow(Vert.getValue()-1);
	return true;
      }
    }

    if (e.id == Event.KEY_ACTION_RELEASE)
    {
      if (e.key == Event.DOWN)
      {
	int value;
   
        if (Selection.showHighlighter() == false)
          return true;

        if (Selection.highlighterDown() == false)
        {
	  value = Vert.getValue();
	  if (value < Vert.getMaximum())
	  {
	    Vert.setValue(value+1);
	    Selection.setTopRow(value+1);
	    AdjustScrollbars();
	  }
          Selection.highlighterDown();
        }

	return true;
      }

      if (e.key == Event.UP)
      {
	int value;
   
        if (Selection.showHighlighter() == false)
          return true;

        if (Selection.highlighterUp() == false)
        {
	  value = Vert.getValue();
	  if (value > Vert.getMinimum())
	  {
	    Vert.setValue(value-1);
	    Selection.setTopRow(value-1);
	    AdjustScrollbars();
	  }
          Selection.highlighterUp();
        }

	return true;
      }

      if (e.key == Event.LEFT)
      {
	int value;
   
	value = Horiz.getValue();
	if (value > Horiz.getMinimum())
	{
	  Horiz.setValue(value-1);
	  Selection.setLeftCol(value-1);
	  AdjustScrollbars();
	}
	return true;
      }

      if (e.key == Event.RIGHT)
      {
	int value;
   
	value = Horiz.getValue();
	if (value < Horiz.getMaximum())
	{
	  Horiz.setValue(value+1);
	  Selection.setLeftCol(value+1);
	  AdjustScrollbars();
	}
	return true;
      }

      if (e.key == Event.PGUP)
      {
	int value;
   
	value = Vert.getValue();
	if (value > Vert.getMinimum())
	{
          value = value - Selection.getVisibleRows();
          if (value < Vert.getMinimum())
            value = Vert.getMinimum();
	  Vert.setValue(value);
	  Selection.setTopRow(value);
	  AdjustScrollbars();
	}
//	Selection.showHighlighter();

	return true;
      }

      if (e.key == Event.PGDN)
      {
	int value;
   
	value = Vert.getValue();
	if (value < Vert.getMaximum())
	{
          value = value + Selection.getVisibleRows();
          if (value >= Vert.getMaximum())
            return true;
	  Vert.setValue(value);
	  Selection.setTopRow(value);
	  AdjustScrollbars();
	}
//	Selection.showHighlighter();

	return true;
      }

    }

    if (e.id == Event.KEY_RELEASE)
    {
      if (e.key == ' ')
      {
        Selection.selectHighlighted();
        return true;
      }
      if (e.key == '\n')
      {
        int row;

        row = Selection.getHighlighted();

        if (row != -1)
	  postEvent(new Event(this, Event.ACTION_EVENT, new Integer(row)));
      }
    }

    return super.handleEvent(e);
  }

  /** Sets the number of columns in this Table.  The leftmost column is
      number zero. */
  public void setNumColumns(int c)
  {
    Horiz.setValues(0, 1, 0, c-1);
    Selection.setNumColumns(c);
  }

  /** Sets the width of a columns in Pixels. */
  public void setColWidth(int c, int w)
  {
    Selection.setColWidth(c,w);
  }

  /** Sets the headers of the columns. */
  public void setHeaders(String[] s)
  {
    Selection.setHeaders(s);
  }

  /** Sets the headers of the columns. */
  public void setHeaders(Vector v)
  {
    String[] s;
    int i;
  
    s = new String[v.size()];
    
    for (i=0; i<v.size(); i++)
      s[i] = (String) v.elementAt(i);

    Selection.setHeaders(s);
  }

  /** Adds a row to the bottom of the Table */
  public void addRow(Vector v)
  {
    Vert.setValues(0, 1, 0, Vert.getMaximum()+1);
    Selection.addRow(v);
  }

  /** Adds a row at the specified location. */
  public void addRow(Vector v, int row)
  {
    Vert.setValues(0, 1, 0, Vert.getMaximum()+1);
    Selection.addRow(v, row);
  }

  /** Deletes all of the rows in the Table */
  public void deleteAll()
  {
    Vert.setValues(0, 1, 0, 0);
    Selection.deleteAll();
  }

  /** Deletes the specified row */
  public void deleteRow(int row)
  {
    if (row >= 0 && row < Vert.getMaximum())
    {
      Vert.setValues(0, 1, 0, Vert.getMaximum()-1);
      Selection.deleteRow(row);
    }
  }

  /** Returns the number of rows in the Table */
  public int countRows()
  {
    return Selection.countRows();
  }

  /** Deselect a row */
  public void deselect(int row)
  {
    Selection.selectRow(row, false);
  }

  /** Select a row */
  public void select(int row)
  {
    Selection.selectRow(row, true);
  }

  /** Return the contents of a row */
  public Vector getRow(int row)
  {
    return Selection.getRow(row);
  }

  /** Return the number of rows visible */
  public int getVisibleRows()
  {
    return Selection.getVisibleRows();
  }

  /** Return the number of columns visible */
  public int getVisibleCols()
  {
    return Selection.getVisibleCols();
  }

  /** Return the index of the topmost selected item, 
      returns -1 if no item is selected */
  public int getSelectedIndex()
  {
    return Selection.getSelectedIndex();
  }

  /** Return the selected indexes */
  public int[] getSelectedIndexes()
  {
    return Selection.getSelectedIndexes();
  }

  /** Returns the first selected row, or null if no row is selected. */
  public Vector getSelectedRow()
  {
    return Selection.getSelectedRow();
  }

  /** Returns a Vector of Vectors containing row data. */
  public Vector getSelectedRows()
  {
    return Selection.getSelectedRows();
  }

  /** Returns true if the specified row is selected */
  public boolean isSelected(int row)
  {
    return Selection.isSelected(row);
  }

  /** Set the Font used by the Table */
  public void setFont(Font f)
  {
    Selection.setFont(f);
  }

  /** If true, multiple selections are enabled */
  public void setMultipleSelections(boolean m)
  {
    Selection.setMultipleSelections(m);
  }

  private void AdjustScrollbars()
  {
    int cols, rows;
    int value, min, max;
  
    cols = Selection.getVisibleCols();
    rows = Selection.getVisibleRows();

    Horiz.setPageIncrement(cols);
    Vert.setPageIncrement(rows);
  }

  public void repaint()
  {
//    super.repaint();
    Selection.repaint();
/*    Vert.repaint(); 
    Horiz.repaint(); */
  }

  public synchronized void show()
  {
    super.show();
    AdjustScrollbars();
  }
}
