package rossi.Pop;

import java.lang.*;
import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import rossi.Component.*;

class AdminYesNoDialog extends Frame
{
  Button YesButton, NoButton, CancelButton;
  boolean Answer, Canceled;
  boolean Disposed = false;
  Frame Parent;

  AdminYesNoDialog(Frame Parent, String Text, boolean HaveCancel)
  {
    //super(Parent, true);

    this.Parent = Parent;
    Panel ButtonPanel;
    Label TextLabel;

    setLayout(new BorderLayout(20,20)); 
    TextLabel = new Label(Text);
    TextLabel.setFont(new Font("TimesRoman", Font.BOLD, 24));
    add("North", TextLabel);
 
    ButtonPanel = new Panel();
    add("South", ButtonPanel);

    ButtonPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 10));
    YesButton = new Button("Yes");
    NoButton = new Button("No");
    if (HaveCancel)
      CancelButton = new Button("Cancel");
    ButtonPanel.add(YesButton);
    ButtonPanel.add(NoButton);
    if (HaveCancel)
      ButtonPanel.add(CancelButton);
  }
 
  public synchronized void show() 
  {
    Rectangle bounds = Parent.bounds();
    Rectangle abounds = bounds();

    move(bounds.x + (bounds.width - abounds.width)/ 2,
         bounds.y + (bounds.height - abounds.height)/2);

    super.show();
  }


  public boolean action(Event e, Object arg)
  {
    Event evt = null;
    Component Parent = this.Parent;


    if (e.target == YesButton)
    {
      evt = new Event(this, Event.ACTION_EVENT, "Yes");
      Answer = true;
      Canceled = false;
    }

    if (e.target == NoButton)
    {
      evt = new Event(this, Event.ACTION_EVENT, "No");
      Answer = false;
      Canceled = false;
    }

    if (e.target == CancelButton)
    {
      evt = new Event(this, Event.ACTION_EVENT, "Cancel");
      Answer = false;
      Canceled = true;
    }

    if (evt != null)
    {
      while(Parent.handleEvent(evt) == false)
      {
	Parent = Parent.getParent();
	if (Parent == null)
	  return false;
      } 
      hide();
      dispose();
      Disposed = true;
      return true;
    }
    return false;
  }
}

class AdminErrorDialog extends Frame
{
  Button YesButton;
  boolean Answer;
  boolean Disposed = false;
  Frame Parent;

  AdminErrorDialog(Frame Parent, String Text)
  {
    //super(Parent, true);

    Panel ButtonPanel;
    Label TextLabel;

    setTitle(Text);
    setLayout(new BorderLayout(20,20)); 
    TextLabel = new Label(Text);
    TextLabel.setFont(new Font("TimesRoman", Font.BOLD, 24));
    add("North", TextLabel);
 
    ButtonPanel = new Panel();
    add("South", ButtonPanel);

    ButtonPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 10));
    YesButton = new Button("Ok");
    ButtonPanel.add(YesButton);
    this.Parent = Parent;
    pack();
    show();
  }
 
  public synchronized void show() 
  {
    Rectangle bounds = Parent.bounds();
    Rectangle abounds = bounds();

    move(bounds.x + (bounds.width - abounds.width)/ 2,
         bounds.y + (bounds.height - abounds.height)/2);

    super.show();
  }


  public boolean action(Event e, Object arg)
  {
    if (e.target == YesButton)
    {
      hide();
      dispose();
      Disposed = true;
      return true;
    }

    return false;
  }
}

class AdminMessageViewer extends Frame implements Runnable
{
  TextArea t;
  DataInputStream din = null;
  Thread ReaderThread;

  AdminMessageViewer(InputStream in)
  {
    t = new TextArea(20,80);
    t.setEditable(false);
    t.setFont(new Font("Courier", Font.PLAIN, 12));
    setLayout(new BorderLayout());
    add("Center", t);
    pack();
    show();
    ReaderThread = new Thread(this);
    din = new DataInputStream(in);
    ReaderThread.start();
  }

  public void run()
  {
    String Line;
    String Data;
    int linecnt;

    try
    {
      linecnt = 0;
      Data = "";
      while (true)
      {
        Line = din.readLine(); 
        if (Line == null)
          break;
        if (Line.startsWith("From: "))
          setTitle("Mail From: "+Line.substring(5));
        Data = Data+Line+"\n";
        linecnt++;
        if (linecnt == 100)
        {
          t.appendText(Data);
          Data = "";
          linecnt = 0;
        }
      }
      t.appendText(Data);
      din.close();
      din = null;
    }
    catch (IOException e)
    {
    }
  }

  public boolean handleEvent(Event e)
  {
    if (e.id == Event.WINDOW_DESTROY)
    {
      hide();
      dispose();
      ReaderThread.stop();
      try
      {
        if (din != null)
          din.close();
      }
      catch (IOException excpt)
      {
      }
      return true;
    }
    
    return super.handleEvent(e);
  }
}

public class Admin extends Frame implements Runnable
{
  Configuration Config;
  Properties ServerProps;
  MailFactory MailDrop;
  Table MsgTable;
  TextField UserNameText;
  Button OkButton;
  MenuBar AdminMenuBar;
  Menu MessageMenu;
  MenuItem ViewMessage;
  MenuItem DeleteMessage;
  AdminYesNoDialog DeleteConfirm;
  Image UnreadImage;
  String CurrentUser;

  public Admin()
  {
    File PropFile;
    FileInputStream PropIn;
    Panel TopPanel;
    String TableHeaders[] = {"#", "", "Size", "From", "Subject"};

    /* Get the properties */
    PropFile = Server.LoadProperties();

    ServerProps = new Properties();
    try
    {
      PropIn = new FileInputStream(PropFile);
      ServerProps.load(PropIn);
      PropIn.close();
    }
    catch (IOException e)
    {
    }

    /* Get the user list */
    Config = new Configuration(ServerProps.getProperty("rossi.Pop.UserList"));

    /* Get a new MailFactory */
    MailDrop = newMailFactory(ServerProps.getProperty("rossi.Pop.MailFactory"));
    MailDrop.setProperties(ServerProps);

    /* Setup the GUI stuff */

    setBackground(Color.lightGray);
    setLayout(new BorderLayout());
    TopPanel = new Panel();
    add("North", TopPanel);
    TopPanel.setLayout(new FlowLayout());
    TopPanel.add(new Label("UserName"));
    UserNameText = new TextField(20);
    TopPanel.add(UserNameText);
    OkButton = new Button("Ok");
    TopPanel.add(OkButton);

    /* Setup the message table */    
    MsgTable = new Table();
    MsgTable.setFont(new Font("TimesNewRoman", Font.PLAIN, 12));
    MsgTable.setNumColumns(5);
    MsgTable.setHeaders(TableHeaders);
    MsgTable.setColWidth(0, 40);
    MsgTable.setColWidth(1, 25);
    MsgTable.setColWidth(2, 40);
    MsgTable.setColWidth(3, 200);
    MsgTable.setColWidth(4, 1000);
    MsgTable.setMultipleSelections(true);
    add("Center", MsgTable);


    /* Do the menus */

    AdminMenuBar = new MenuBar();
    setMenuBar(AdminMenuBar);
    MessageMenu = new Menu("Message");
    AdminMenuBar.add(MessageMenu);
    ViewMessage = new MenuItem("View");
    DeleteMessage = new MenuItem("Delete");
    MessageMenu.add(ViewMessage);
    MessageMenu.add(DeleteMessage);
    ViewMessage.disable();
    DeleteMessage.disable();

    setTitle("Rossi Engineering's J-POP Server -- Admin Screen");

    resize(500,300);
    show();
    getUnreadImage();
  }

  private void getUnreadImage()
  {
    Graphics g;

    UnreadImage = createImage(21,11);
    g=UnreadImage.getGraphics();
    g.setColor(Color.white);
    g.fillRect(0,0,21,11);
    g.setColor(Color.black);
    g.drawRect(0,0,20,10);
    g.drawLine(0,0,10,5);
    g.drawLine(20,0,10,5);
    g.drawLine(0,10, 4, 4);
    g.drawLine(20,10, 16, 4);
  }

  private MailFactory newMailFactory(String MailFactoryName)
  {
    Class c;
    Class Interfaces[];
    MailFactory m;
    boolean IsMailFactory = false;
    int i;

    if (MailFactoryName == null)
    {
      System.err.println("No mail factory class specified.");
      return null;
    }

    try
    {
      c = Class.forName(MailFactoryName);
    }
    catch (ClassNotFoundException e)
    {
      System.err.println("Class "+MailFactoryName+" not found");
      return null;
    }

    Interfaces = c.getInterfaces();
  
    for (i=0; i<Interfaces.length; i++)
    {
      if (Interfaces[i].getName().equals("rossi.Pop.MailFactory"))
        IsMailFactory = true;
    }

    if (IsMailFactory == false)
    {
      System.err.println("Class "+MailFactoryName+" is not a MailFactory");
      return null;
    }

    try
    {
      m = (MailFactory) c.newInstance();
    }
    catch (InstantiationException e)
    {
      return null;
    }
    catch (IllegalAccessException e)
    {
      return null;
    }

    return m;
  }

  private void LoadTable()
  {
    String UserName;
    MailSpec[] Mail;
    UserSpec User;
    boolean UserFound = false;
    Vector v;
    DataInputStream MailIn;
    String From="", Subject="";
    String Line;
    int i;

    /* Get the user name */
    UserName = UserNameText.getText();
    CurrentUser = UserName;

    /* Search the user list for the spec. user */

    for (i=0; i<Config.UserList.size(); i++)
    {
       User = (UserSpec) Config.UserList.elementAt(i);
       if (User.Name.equals(UserName))
       {
	   if (MailDrop != null)
	   {
	     if (MailDrop.setMailbox(User.Fname))
	       UserFound = true;
	   }
	   break;
       }
    }

    if (UserFound == false)
    {
      new AdminErrorDialog(this, "Unknown User");
      return;
    }

    /* User was found, so start loading the table */

    MsgTable.deleteAll();  

    Mail = MailDrop.getMailList();
    
    for (i=0; i<Mail.length; i++)
    {
      v = new Vector();
      v.addElement((new Integer(i+1)).toString());
      if (Mail[i].Read)
        v.addElement("");
      else
        v.addElement(UnreadImage);
      v.addElement((Mail[i].Size/1024)+"K");

      From = "";
      Subject = "";
      MailIn = new DataInputStream(MailDrop.getMessageInputStream(i));
      try
      {
        while (true)
        {
          Line = MailIn.readLine();
          if (Line == null)
            break;

          if (Line.equals(""))
            break;

          if (Line.startsWith("From: "))
            From = Line.substring(5);

          if (Line.startsWith("Subject: "))
            Subject = Line.substring(8);
        }

        MailIn.close();
      }
      catch (IOException e)
      {
      }
      v.addElement(From);
      v.addElement(Subject);
      MsgTable.addRow(v);

      if ( (i%20) == 19)
        MsgTable.repaint();
    }
    MsgTable.repaint();
    ViewMessage.enable();
    DeleteMessage.enable();
  }

  public void run()
  {
    ViewMessage.disable();
    DeleteMessage.disable();
    MailDrop.updateMaildrop();
    setCursor(Frame.WAIT_CURSOR);
    LoadTable();
    setCursor(Frame.DEFAULT_CURSOR);
    ViewMessage.enable();
    DeleteMessage.enable();
  }

  public boolean action(Event evt, Object arg)
  {
    if (evt.target == OkButton)
    {
      setCursor(Frame.WAIT_CURSOR);
      LoadTable();
      setCursor(Frame.DEFAULT_CURSOR);
      return true;
    }

    if (evt.target == MsgTable)
    {
      if (ViewMessage.isEnabled())
        new AdminMessageViewer(MailDrop.getMessageInputStream(((Integer)(arg)).intValue()));
      return true;
    }

    if (evt.target == ViewMessage)
    {
      int Sel;

      Sel = MsgTable.getSelectedIndex();

      if (Sel != -1)
        new AdminMessageViewer(MailDrop.getMessageInputStream(Sel));
      return true;
    }

    if (evt.target == DeleteMessage)
    {
      DeleteConfirm = new AdminYesNoDialog(this, 
          "Delete "+MsgTable.getSelectedIndexes().length+" messages?", false);
      DeleteConfirm.pack();
      DeleteConfirm.show();
      return true;
    }

    if (evt.target == DeleteConfirm)
    {
      int i;
      int[] Sel;
      MailSpec[] Mail;

      UserNameText.setText(CurrentUser);
      Mail = MailDrop.getMailList();

      Sel = MsgTable.getSelectedIndexes();

      if (DeleteConfirm.Answer)
      {
        for (i=0; i<Sel.length; i++)
        {
          System.out.println(Sel[i]);
          Mail[Sel[i]].Delete = true;  
        }
        (new Thread(this)).start();
      }
      return true;
    } 

    return false;
  }

  public boolean handleEvent(Event e)
  {
    if (e.id == Event.WINDOW_DESTROY)
    {
      hide();
      if (DeleteMessage.isEnabled())
        MailDrop.updateMaildrop();
      dispose();
      System.exit(1);
      return true;
    }

    if (e.target == UserNameText && e.id == Event.KEY_ACTION_RELEASE && e.key=='\n')
    {
      LoadTable();
      return true;
    }
    
    return super.handleEvent(e);
  }


  public static void main(String args[])
  {
    new Admin();
  }
}
