
package rossi.Pop;

/* RSA Data Security, Inc. MD5 Message-Digest Algorithm              */
/* Ported to Java from the C source contained in RFC 1321 by W.Rossi */
/* Ported on Jan. 6, 1997                                            */
/*                                                                   */
/* Test suite is contained here, invoked by function main().  Should */
/* produce the following output:                                     */
/*                                                                   */
/* MD5 test suite:
   MD5 ("") = d41d8cd98f00b204e9800998ecf8427e
   MD5 ("a") = 0cc175b9c0f1b6a831c399e269772661
   MD5 ("abc") = 900150983cd24fb0d6963f7d28e17f72
   MD5 ("message digest") = f96b697d7cb7938d525a2f31aaf161d0
   MD5 ("abcdefghijklmnopqrstuvwxyz") = c3fcd3d76192e4007dfb496cca67e13b
   MD5 ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789") = d174ab98d277d9f5a5611c2c9f419d9f
   MD5 ("12345678901234567890123456789012345678901234567890123456789012345678901234567890") = 57edf4a22be3c955ac49da2e2107b67a
*/

import java.lang.*;


/* MD5 context. */
class MD5_CTX
{
  int state[] = new int[4];        /* state (ABCD) */
  int count[] = new int[2];        /* number of bits, modulo 2^64 (lsb first) */
  byte buffer[] = new byte[64];    /* input buffer */
};

public class MD5
{
  private static int S11=7;
  private static int S12=12;
  private static int S13=17;
  private static int S14=22;
  private static int S21=5;
  private static int S22=9;
  private static int S23=14;
  private static int S24=20;
  private static int S31=4;
  private static int S32=11;
  private static int S33=16;
  private static int S34=23;
  private static int S41=6;
  private static int S42=10;
  private static int S43=15;
  private static int S44=21;


  private static byte PADDING[] = {
  (byte) 0x80, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
  (byte) 0 };


  /* F, G, H and I are basic MD5 functions.
   */

  private static int F(int x, int y, int z)
  {
    return (((x) & (y)) | ((~x) & (z)));
  }

  private static int G(int x, int y, int z)
  {
    return (((x) & (z)) | ((y) & (~z)));
  }

  private static int H(int x, int y, int z)
  {
    return ((x) ^ (y) ^ (z));
  }

  private static int I(int x, int y, int z)
  {
    return ((y) ^ ((x) | (~z)));
  }


  /* ROTATE_LEFT rotates x left n bits.
   */
  private static int ROTATE_LEFT(int x, int n)
  {
    return (((x) << (n)) | ((x) >>> (32-(n))));
  }

  /* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
  Rotation is separate from addition to prevent recomputation.
   */

  private static int FF(int a, int b, int c, int d, int x, int s, int ac)
  { 
    a += F ((b), (c), (d)) + (x) + (int)(ac); 
    a = ROTATE_LEFT ((a), (s)); 
    a += (b); 
    return a;
  }

  private static int GG(int a, int b, int c, int d, int x, int s, int ac)
  { 
    a += G ((b), (c), (d)) + (x) + (int)(ac); 
    a = ROTATE_LEFT ((a), (s)); 
    a += (b); 
    return a;
  }

  private static int HH(int a, int b, int c, int d, int x, int s, int ac)
  { 
    a += H ((b), (c), (d)) + (x) + (int)(ac); 
    a = ROTATE_LEFT ((a), (s)); 
    a += (b); 
    return a;
  }

  private static int II(int a, int b, int c, int d, int x, int s, int ac)
  { 
    a += I ((b), (c), (d)) + (x) + (int)(ac); 
    a = ROTATE_LEFT ((a), (s)); 
    a += (b); 
    return a;
  }

  /* MD5 initialization. Begins an MD5 operation, writing a new context.
   */
  public static void MD5Init (MD5_CTX context)
  {
    context.count[0] = context.count[1] = 0;
    /* Load magic initialization constants.
  */
    context.state[0] = 0x67452301;
    context.state[1] = 0xefcdab89;
    context.state[2] = 0x98badcfe;
    context.state[3] = 0x10325476;
  }

  /* MD5 block update operation. Continues an MD5 message-digest
    operation, processing another message block, and updating the
    context.
   */
  public static void MD5Update (MD5_CTX context, byte[] input, int inputLen)
  {
    int i, index, partLen;

    /* Compute number of bytes mod 64 */
    index = (int)((context.count[0] >>> 3) & 0x3F);

    /* Update number of bits */
    if ((context.count[0] += ((int)inputLen << 3)) < ((int)inputLen << 3))
      context.count[1]++;
    context.count[1] += ((int)inputLen >> 29);

    partLen = 64 - index;

    /* Transform as many times as possible.
  */
    if (inputLen >= partLen)
    {
      MD5_memcpy (context.buffer, index, input, 0, partLen);
      MD5Transform (context.state, context.buffer, 0);

      for (i = partLen; i + 63 < inputLen; i += 64)
        MD5Transform (context.state, input, i);

      index = 0;
    }
    else
      i = 0;

    /* Buffer remaining input */
    MD5_memcpy(context.buffer, index, input, i, inputLen-i);
  }

  /* MD5 finalization. Ends an MD5 message-digest operation, writing the
    the message digest and zeroizing the context.
   */
  public static void MD5Final (byte[] digest, MD5_CTX context)
  {
    byte bits[] = new byte[8];
    int index, padLen;

    /* Save number of bits */
    Encode (bits, context.count, 8);

    /* Pad out to 56 mod 64.
  */
    index = (int)((context.count[0] >>> 3) & 0x3f);
    padLen = (index < 56) ? (56 - index) : (120 - index);
    MD5Update (context, PADDING, padLen);

    /* Append length (before padding) */
    MD5Update (context, bits, 8);

    /* Store state in digest */
    Encode (digest, context.state, 16);

    /* Zeroize sensitive information.
  */
//    MD5_memset ((POINTER)context, 0, sizeof (*context));
  }

  /* MD5 basic transformation. Transforms state based on block.
   */
  private static void MD5Transform (int[] state, byte[] block, int blkindex)
  {
    int a = state[0], b = state[1], c = state[2], d = state[3], x[] = new int[16];

    Decode (x, block, blkindex, 64);

    /* Round 1 */
    a = FF (a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
    d = FF (d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
    c = FF (c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
    b = FF (b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
    a = FF (a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
    d = FF (d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
    c = FF (c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
    b = FF (b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
    a = FF (a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
    d = FF (d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
    c = FF (c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
    b = FF (b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
    a = FF (a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
    d = FF (d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
    c = FF (c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
    b = FF (b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

   /* Round 2 */
    a = GG (a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
    d = GG (d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
    c = GG (c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
    b = GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
    a = GG (a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
    d = GG (d, a, b, c, x[10], S22,  0x2441453); /* 22 */
    c = GG (c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
    b = GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
    a = GG (a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
    d = GG (d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
    c = GG (c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */


  /*
  Rivest                                                         [Page 13]
  
  RFC 1321              MD5 Message-Digest Algorithm            April 1992
  */

    b = GG (b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
    a = GG (a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
    d = GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
    c = GG (c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
    b = GG (b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

    /* Round 3 */
    a = HH (a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
    d = HH (d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
    c = HH (c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
    b = HH (b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
    a = HH (a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
    d = HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
    c = HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
    b = HH (b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
    a = HH (a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
    d = HH (d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
    c = HH (c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
    b = HH (b, c, d, a, x[ 6], S34,  0x4881d05); /* 44 */
    a = HH (a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
    d = HH (d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
    c = HH (c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
    b = HH (b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

    /* Round 4 */
    a = II (a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
    d = II (d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
    c = II (c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
    b = II (b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
    a = II (a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
    d = II (d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
    c = II (c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
    b = II (b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
    a = II (a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
    d = II (d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
    c = II (c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
    b = II (b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
    a = II (a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
    d = II (d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
    c = II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
    b = II (b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;

    /* Zeroize sensitive information.
  */
//    MD5_memset ((POINTER)x, 0, sizeof (x));
  }

  /* Encodes input (UINT4) into output (unsigned char). Assumes len is
    a multiple of 4.
   */
  private static void Encode (byte[] output, int[] input, int len)
  {
    int i, j;

    for (i = 0, j = 0; j < len; i++, j += 4) {
      output[j] = (byte)(input[i] & 0xff);
      output[j+1] = (byte)((input[i] >>> 8) & 0xff);
      output[j+2] = (byte)((input[i] >>> 16) & 0xff);
      output[j+3] = (byte)((input[i] >>> 24) & 0xff);
    }
  }

  /* Decodes input (unsigned char) into output (UINT4). Assumes len is
    a multiple of 4.
   */
  private static void Decode (int[] output, byte[] input, int inputindex, int len)
  {
    int i, j;

    for (i = 0, j = 0; j < len; i++, j += 4)
      output[i] = ((int)input[j+inputindex]&0xff) | (((int)input[j+1+inputindex]&0xff) << 8) |
      (((int)input[j+2+inputindex]&0xff) << 16) | (((int)input[j+3+inputindex]&0xff) << 24);
  }

  /* Note: Replace "for loop" with standard memcpy if possible.
   */

  private static void MD5_memcpy (byte[] output, int outindex, 
                                  byte[] input, int inindex, int len)
  {
    int i;

    for (i = 0; i < len; i++)
      output[i+outindex] = input[i+inindex];
  }

  /* Note: Replace "for loop" with standard memset if possible.
   */
/*
  static void MD5_memset (output, value, len)
  POINTER output;
  int value;
  unsigned int len;
  {
    unsigned int i;

    for (i = 0; i < len; i++)
   ((char *)output)[i] = (char)value;
  }
*/

  /* Code from here down is for the test suite. */

  /* Digests a reference suite of strings and prints the results.
   */
  private static void MDTestSuite ()
  {
    System.out.println ("MD5 test suite:");

    MDString ("");
    MDString ("a");
    MDString ("abc");
    MDString ("message digest");
    MDString ("abcdefghijklmnopqrstuvwxyz");
    MDString
   ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789");
    MDString ("12345678901234567890123456789012345678901234567890123456789012345678901234567890");
  }


  /* Digests a string and prints the result.
   */
  private static void MDString (String string)
  {
    MD5_CTX context = new MD5_CTX();
    byte digest[] = new byte[16];
    int len = string.length();
    byte ByteInput[] = new byte[len];

    string.getBytes(0,len,ByteInput,0);

    MD5Init (context);
    MD5Update (context, ByteInput, len);
    MD5Final (digest, context);

    System.out.print ("MD5 (\""+string+"\") = ");
    MDPrint (digest);
    System.out.print ("\n");
  }


  private static void MDPrint (byte[] digest)
  {
    int i;
    String s;

    for (i = 0; i < 16; i++)
    {
      s = Integer.toHexString((int)digest[i]&0x000000ff);
      if (s.length() == 1)
        s = "0"+s;
      System.out.print(s);
    }
  }

  public static void main(String args[])
  {
    MDTestSuite();
  }
};

