
package rossi.Pop;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

/** Interface that represents a mail drop */

public interface MailFactory
{
  /** Returns information on mail available in the maildrop */

  public MailSpec[] getMailList();

  /** Sets the name of the mailbox file or directory.  This must
      be called before any of the other methods.  Returns true
      if successfully opened mailbox, false otherwise. */

  public boolean setMailbox(String Fname);

  /** Returns an InputStream for reading actual message data */
  public InputStream getMessageInputStream(int MsgNum);

  /** Updates the maildrop, marks messages as read, deletes 
      messages that need deleting. */

  public void updateMaildrop();

  /** Sets the Properties for this MailFactory.  This is used so that
      users of the MailFactory object can specify MailFactory dependant
      options.  If the specified properties is null, then the MailFactory
      should use defaults */

  public void setProperties(Properties p);

  /** Returns a list of Property names understood by this MailFactory */
  public String[] listPropertyNames();

  /** Returns a Properties object initialized with default Properties */
  public Properties getDefaultProperties();

  /** Verifys the correctness of the Properties set wit setProperties().
      returns a String with an error message if there is a problem, and
      returns null if all went well */

  public String checkProperties();
}
