
package rossi.Pop;

import java.lang.*;
import java.io.*;
import java.util.*;

/** This class represents the POP3 protocol.  MailFactory classes handle
    interaction with the mail locally.  The code using this class is 
    responsible for all networking stuff **/

public class PopProtocol implements Runnable
{
  private static final int AUTHORIZATION = 0;
  private static final int TRANSACTION = 1;
  private static final int UPDATE = 2;
  private static final int EXIT = 3;
  private static final String PositiveResponse = "+OK";
  private static final String NegativeResponse = "-ERR";
  private static final Vector CurrentUsers = new Vector();
  private boolean LoggedIn = false;
  private int State = AUTHORIZATION;
  private InputStream in;
  private OutputStream out;
  private Configuration Config;
  private DataInputStream din;
  private PrintStream pout;
  private PrintStream log;
  private String CmdLine = null;
  private String UserName = null;
  private UserSpec User;
  private MailFactory MailDrop = null;
  private String MailFactoryName;
  private String BannerTimeStamp;
  private String HostName;
  private String SessionIdentifier;
  private boolean EnableLAST = true;
  private boolean EnableUIDL = true;
  private boolean EnableTOP = true;
  private boolean EnableAPOP = true;
  private boolean EnableUSERPASS = true;
  private Properties MailFactoryProperties = null;

  public PopProtocol(InputStream in, OutputStream out, Configuration Config, String HostName)
  {
    this.in = in;
    this.out = out;
    this.Config = Config;
    this.HostName = HostName;

    din = new DataInputStream(in);
    pout = new PrintStream(out, true);
    SessionIdentifier = Integer.toHexString(hashCode());
  }

  public void run()
  {
    StringTokenizer st = null;

    if (EnableAPOP)
      BannerTimeStamp = "<"+System.currentTimeMillis()+"@"+HostName+">";
    else
      BannerTimeStamp = "";

    /* Output the greeting */
    this.println(PositiveResponse+" POP3 server ready "+BannerTimeStamp+"\r");

    while (true)
    {
      if (State != UPDATE)
      {
	try
	{
	  CmdLine = din.readLine(); 
          synchronized(this) {notifyAll();}
	}
	catch (IOException e)
	{
          CmdLine = null;
	}

	if (CmdLine == null)
	  break;

        /* Log incoming commands */
	if (log != null)
          log.println(SessionIdentifier+" > "+CmdLine+"\r");

	st = new StringTokenizer(CmdLine, " \r\n");
      }

      switch (State)
      {
        
        case AUTHORIZATION:
          AuthorizationState(st);
          break;

        case TRANSACTION:
          TransactionState(st);
          break;

        case UPDATE:
          UpdateState();
          break;
      }

      if (State == EXIT)
        break;
    }
   
    try
    {
      in.close();
      out.close();
    }
    catch (IOException e)
    {
    }

    /* Remove the user from the list of those logged in */
    if (LoggedIn)
      LogoutUser();
  }

  boolean CheckLoggedIn()
  {
    long start = System.currentTimeMillis();

    synchronized(CurrentUsers)
    {
      /* Wait up to 25 seconds for another login to clear */
      while ( (System.currentTimeMillis() - start) < 25000)
      {
        if (internalCheckLoggedIn() == false)
          return false;

        try
        {
          CurrentUsers.wait(5000);
        }
        catch (InterruptedException ex)
        {
        }
      }
    }

    return true;
  }

  boolean internalCheckLoggedIn()
  {
    synchronized(CurrentUsers)
    {
      boolean LoggedIn = false;
      int i;
  
      if (UserName == null)
	return false;
  
      for (i=0; i<CurrentUsers.size(); i++)
      {
	if (UserName.equalsIgnoreCase((String) CurrentUsers.elementAt(i)))
	{
	  LoggedIn = true;
	  break;
	}
      }
  
      /* If the user wasn't already logged in, add their name to the list */
  
      if (LoggedIn == false) 
	CurrentUsers.addElement(UserName);
  
      return LoggedIn;
    }
  }

  void LogoutUser()
  {
    synchronized(CurrentUsers)
    {
      boolean LoggedIn = false;
      int i;
  
      if (UserName == null)
	return;
  
      for (i=0; i<CurrentUsers.size(); i++)
      {
	if (UserName.equalsIgnoreCase((String) CurrentUsers.elementAt(i)))
	{
	  CurrentUsers.removeElementAt(i);
	  break;
	}
      }

      CurrentUsers.notify();
    }
  }

  void AuthorizationState(StringTokenizer st) 
  {
    String Cmd;
    int Count;
    int i, j;
    
    Count = st.countTokens();

    if (Count < 1)
    {
      this.println(NegativeResponse+" Enter USER or PASS\r");
      return;
    }

    Cmd = st.nextToken();
    Cmd = Cmd.toUpperCase();
    
    if (Cmd.equals("USER") && Count > 1 && EnableUSERPASS == true)
    {
      UserName = new String(st.nextToken());
      this.println(PositiveResponse+"\r");
      return;
    }

    if (Cmd.equals("PASS") && Count > 1 && EnableUSERPASS == true)
    {
      String Pass;
 
      Pass = st.nextToken();

      /* Search the user list for the spec. user */

      for (i=0; i<Config.UserList.size(); i++)
      {
         User = (UserSpec) Config.UserList.elementAt(i);
         if (User.Name.equals(UserName))
         {
           if (User.Password.equals(Pass) && !CheckLoggedIn())
           {
             System.err.println(UserName+" logged in via USER|PASS");
             MailDrop = newMailFactory();
             if (MailDrop != null)
             {
               MailDrop.setProperties(MailFactoryProperties);
               if (MailDrop.setMailbox(User.Fname))
                 State = TRANSACTION;
             }
             break;
           }
         }
      }

      if (State == AUTHORIZATION)
      {
        this.println(NegativeResponse+" Sorry Charlie\r");
        LoggedIn = false;
      }
      else
      {
        this.println(PositiveResponse+" "+User.Name+" Logged in\r");
        logLogin(UserName);
        LoggedIn = true;
      }
      
      return;
    }

    if (Cmd.equals("APOP") && Count > 1 && EnableAPOP == true)
    {
      String Pass;
      MD5_CTX context = new MD5_CTX();
      byte digest[] = new byte[16];
 
      UserName = st.nextToken();
      Pass = st.nextToken();

      /* Search the user list for the spec. user */

      for (i=0; i<Config.UserList.size(); i++)
      {
         User = (UserSpec) Config.UserList.elementAt(i);
         if (User.Name.equals(UserName))
         {
           String ClearTxt, DigestStr = "", s;
           byte[] str;
           int len;
 
           ClearTxt = BannerTimeStamp+User.Password;

           len = ClearTxt.length();
           str = new byte[len];

           ClearTxt.getBytes(0,len,str,0);
	   MD5.MD5Init(context);
	   MD5.MD5Update(context, str, len);
	   MD5.MD5Final(digest, context);
           
	   for (j = 0; j < 16; j++)
	   {
	     s = Integer.toHexString((int)digest[j]&0x000000ff);
	     if (s.length() == 1)
	       s = "0"+s;
	     DigestStr = DigestStr+s;
	   }


           if (Pass.equals(DigestStr) && !CheckLoggedIn())
           {
             System.err.println(UserName+" logged in via APOP");
             MailDrop = newMailFactory();
             if (MailDrop != null)
             {
               MailDrop.setProperties(MailFactoryProperties);
               if (MailDrop.setMailbox(User.Fname))
                 State = TRANSACTION;
             }
             break;
           }
         }
      }

      if (State == AUTHORIZATION)
      {
        this.println(NegativeResponse+" Sorry Charlie\r");
        LoggedIn = false;
      }
      else
      {
        this.println(PositiveResponse+" "+User.Name+" Logged in\r");
        LoggedIn = true;
      }
      
      return;
    }

    if (Cmd.equals("QUIT"))
    { 
      State = EXIT;
      this.println(PositiveResponse+" Better luck next time, hacker\r");
      return;
    }

    this.println(NegativeResponse+" Enter USER or PASS\r");
  }
 
  void TransactionState(StringTokenizer st)
  {
    String Cmd;
    int Count;
    int i;
    
    Count = st.countTokens();

    if (Count < 1)
    {
      this.println(NegativeResponse+" I dont understand that.\r");
      return;
    }

    Cmd = st.nextToken();
    Cmd = Cmd.toUpperCase();
    
    if (Cmd.equals("STAT"))
    {
      ProcessSTAT();
      return;
    }

    if (Cmd.equals("LIST"))
    {
      ProcessLIST(st);
      return;
    }

    if (Cmd.equals("UIDL"))
    {
      ProcessUIDL(st);
      return;
    }

    if (Cmd.equals("RETR") && Count > 1)
    {
      ProcessRETR(st.nextToken());
      return;
    }

    if (Cmd.equals("TOP") && Count > 2)
    {
      ProcessTOP(st);
      return;
    }

    if (Cmd.equals("NOOP"))
    {
      this.println(PositiveResponse+"\r");
      return;
    }

    if (Cmd.equals("DELE") && Count > 1)
    {
      ProcessDELE(st.nextToken());
      return;
    }

    if (Cmd.equals("RSET"))
    {
      ProcessRSET();
      return;
    }

    if (Cmd.equals("LAST"))
    {
      ProcessLAST();
      return;
    }

    if (Cmd.equals("QUIT"))
    { 
      State = UPDATE;
      /* Positive response is set later in UpdateState() */
      return;
    }

    this.println(NegativeResponse+" I don't understand that\r");
  }

  void ProcessSTAT()
  {
    int NumMsg = 0, NumBytes = 0, i;
    MailSpec[] Mail;

    Mail = MailDrop.getMailList();

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Delete == false)
      {
        NumMsg++;
        NumBytes+=Mail[i].Size;
      }
    }
    this.println(PositiveResponse+" "+NumMsg+" "+NumBytes+"\r");
  }

  void ProcessLIST(StringTokenizer st)
  {
    int NumMsg = 0, NumBytes = 0, i = 0;
    MailSpec[] Mail;

    Mail = MailDrop.getMailList();

    if (st.hasMoreTokens())
    {
      try
      {
        i = Integer.parseInt(st.nextToken()); 
      }
      catch (NumberFormatException e)
      {
        this.println(NegativeResponse+" Try a valid number next time\r");
        return;
      }

      i--;

      if (i >= Mail.length || i < 0)
      {
        this.println(NegativeResponse+"\r");
        return;
      }

      if (Mail[i].Delete)
      {
        this.println(NegativeResponse+" Cant LIST a deleted message.\r");
        return;
      }

      this.println(PositiveResponse+" "+(i+1)+" "+Mail[i].Size+"\r");
      return;
    }

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Delete == false)
      {
        NumMsg++;
        NumBytes+=Mail[i].Size;
      }
    }

    this.println(PositiveResponse+" "+NumMsg+" Messages ("+NumBytes+" bytes)\r");

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Delete == false)
      {
        this.println((i+1)+" "+Mail[i].Size+"\r");
      }
    }
    this.println(".\r");

  }

  void ProcessUIDL(StringTokenizer st)
  {
    int NumMsg = 0, NumBytes = 0, i = 0;
    MailSpec[] Mail;

    if (EnableUIDL == false)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    Mail = MailDrop.getMailList();

    if (st.hasMoreTokens())
    {
      try
      {
        i = Integer.parseInt(st.nextToken()); 
      }
      catch (NumberFormatException e)
      {
        this.println(NegativeResponse+" Try a valid number next time\r");
        return;
      }

      i--;

      if (i >= Mail.length || i < 0)
      {
        this.println(NegativeResponse+"\r");
        return;
      }

      if (Mail[i].Delete)
      {
        this.println(NegativeResponse+" Cant UIDL a deleted message\r");
        return;
      }

      this.println(PositiveResponse+" "+(i+1)+" "+Mail[i].UniqID+"\r");
      return;
    }

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Delete == false)
      {
        NumMsg++;
        NumBytes+=Mail[i].Size;
      }
    }

    this.println(PositiveResponse+"\r");

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Delete == false)
      {
        this.println((i+1)+" "+Mail[i].UniqID+"\r");
      }
    }
    this.println(".\r");
  }

  /* This function has been performance optimized.  The
     readLine() and println() functions of the DataInputStream
     and PrintStream classes were found to be terribly slow.
     Since this function moves the bulk of the data, it uses
     the lower level calls read and write. */ 

  void ProcessRETR(String MsgId)
  {
    int NumMsg = 0, NumBytes = 0, i = 0, br;
    InputStream MailIn;
    String Line = null;
    MailSpec[] Mail;

    Mail = MailDrop.getMailList();

    try
    {
      i = Integer.parseInt(MsgId); 
    }
    catch (NumberFormatException e)
    {
      this.println(NegativeResponse+" Try a valid number next time\r");
      return;
    }

    i--;

    if (i >= Mail.length || i < 0)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    if (Mail[i].Delete)
    {
      this.println(NegativeResponse+" Cant RETR a deleted message\r");
      return;
    }


    MailIn = MailDrop.getMessageInputStream(i);

    if (MailIn == null)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    this.println(PositiveResponse+" "+Mail[i].Size+" bytes\r");

    // Flush any output in the PrintStream, as that we are about
    // to bypass it and write directly to the underlying OuputStream.

    pout.flush();  

    try
    {
      byte[] buffer = new byte[1024];
      int j = 0, k, bs;

      /* If the first line of the message start with "From " then
         suppress it from the output */

      bs = 5;
      br = 0;
      while (bs > 0)
      {
        j = MailIn.read(buffer, br, bs);
        if (j == -1)
          break;
        br += j;
        bs -= j;
      }

      if (new String(buffer, 0, 0, 5).equals("From "))
      {
        do
        {
          k = j;
          j = MailIn.read();
          if (j == -1)
            break;
          br++;
        } while (j != '\n');
        if (k != '\r')
          br++;
      }
      else
      {
        br = 0;
        FixEOLandWrite(out, buffer, br);
      }

      while (true)
      {
        j = MailIn.read(buffer);
	if (j == -1)
	  break;

        FixEOLandWrite(out, buffer, j);
      }
     
      /* Ok we have skipped the message separator, now we have
	 to make up the byte count somehow.  Add some whitespace
	 to the end of the next line */

      if (br > 0)
      {
	for (j=0; j<(br-2);j++)
	  out.write(' ');
	out.write('\r');
	out.write('\n');
      }

      out.flush(); // flush the output stream
      MailIn.close();
    }
    catch (IOException e)
    {
    }

    Mail[i].Read = true;
    this.println(".\r");
  }

  private void FixEOLandWrite(OutputStream out, byte[] buffer, int len) throws IOException
  {
    int start=0, segment;
    int i;

    for (i=0; i<len; i++)
    {
      if (buffer[i] != '\n')
        continue;
   
      /* Insert carriage returns if necessary */
      if (i == 0 || buffer[i-1] != '\r')
      {
        out.write(buffer, start, i-start);
        out.write('\r');
        start=i;
      }

      /* Insert a period if neccessary (byte-stuffing) */
      if (i != (len-1) && buffer[i+1] == '.')
      {
        out.write(buffer, start, i-start+1);
        out.write('.');
        start=i+1;
      }
    } 
    out.write(buffer, start, len-start);
  }

  void ProcessTOP(StringTokenizer st)
  {
    int NumMsg = 0, NumBytes = 0, i = 0, NumLines, LineCtr = 0;
    DataInputStream MailIn;
    String Line = null;
    String MsgId, NumLinesStr;
    boolean HeadersDone = false;
    MailSpec[] Mail;

    if (EnableTOP == false)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    Mail = MailDrop.getMailList();

    MsgId = st.nextToken();
    NumLinesStr = st.nextToken();

    try
    {
      i = Integer.parseInt(MsgId); 
      NumLines = Integer.parseInt(NumLinesStr); 
    }
    catch (NumberFormatException e)
    {
      this.println(NegativeResponse+" Try a valid number next time\r");
      return;
    }

    i--;

    if (i >= Mail.length || i < 0 || NumLines < 0)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    if (Mail[i].Delete)
    {
      this.println(NegativeResponse+" Cant TOP a deleted message\r");
      return;
    }

    this.println(PositiveResponse+" "+Mail[i].Size+" bytes\r");

    MailIn = new DataInputStream(MailDrop.getMessageInputStream(i));

    /* Read in the very first line */

    try
    {
      Line = MailIn.readLine();
    }
    catch (IOException e)
    {
    }

    if (Line.regionMatches(0, "From ", 0, 5) == false)
      pout.print(Line+"\r\n");

    while (LineCtr < NumLines || HeadersDone == false)
    {
      try
      {
        Line = MailIn.readLine();
      }
      catch (IOException e)
      {
      }
      if (Line == null)
        break;

      if (HeadersDone)
        LineCtr++;

      if (Line.equals(""))
        HeadersDone = true;

      if (Line.regionMatches(0, ".", 0, 1))
        Line = "."+Line;

      /* Dont send the actual mail data to the logfile */
      pout.print(Line+"\r\n");
    }
  
    try 
    {
      MailIn.close();
    }
    catch (IOException e)
    {
    }

    this.println(".\r");
  }

  void ProcessDELE(String MsgId)
  {
    int i = 0;
    MailSpec[] Mail;

    Mail = MailDrop.getMailList();

    try
    {
      i = Integer.parseInt(MsgId); 
    }
    catch (NumberFormatException e)
    {
      this.println(NegativeResponse+" Try a valid number next time\r");
      return;
    }

    i--;

    if (i >= Mail.length || i < 0)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    if (Mail[i].Delete)
    {
      this.println(NegativeResponse+" Cant DELE a deleted message\r");
      return;
    }

    Mail[i].Delete = true;
    this.println(PositiveResponse+"\r");
  }

  void ProcessRSET()
  {
    int i;
    MailSpec[] Mail;

    Mail = MailDrop.getMailList();

    for (i=0; i<Mail.length; i++)
    {
      Mail[i].Delete = false;
      Mail[i].Read = Mail[i].ReadPreviously;
    }

    this.println(PositiveResponse+"\r");
  }

  void ProcessLAST()
  {
    int i;
    MailSpec[] Mail;
    int LastRead = -1;

    if (EnableLAST == false)
    {
      this.println(NegativeResponse+"\r");
      return;
    }

    Mail = MailDrop.getMailList();

    for (i=0; i<Mail.length; i++)
    {
      if (Mail[i].Read == true && Mail[i].Delete == false)
      {
        LastRead = i;
      }
    }
 
    LastRead++;

    this.println(PositiveResponse+" "+LastRead+"\r");
  }
 
  void UpdateState()
  {
    if (EnableLAST == false)   /* Unmark all read messages */
    {
      MailSpec[] Mail;
      int i;

      Mail = MailDrop.getMailList();

      for (i=0; i<Mail.length; i++)
	Mail[i].Read = false;
    }

    MailDrop.updateMaildrop();

    /* This is the response to the QUIT command that brought us here */
    this.println(PositiveResponse+" Goodbye, pleasure doing business with you\r");

    State = EXIT;
  }

  public void enableLAST()
  {
    EnableLAST=true;
  }

  public void disableLAST()
  {
    EnableLAST=false;
  }

  public void enableUIDL()
  {
    EnableUIDL=true;
  }

  public void disableUIDL()
  {
    EnableUIDL=false;
  }

  public void enableTOP()
  {
    EnableTOP=true;
  }

  public void disableTOP()
  {
    EnableTOP=false;
  }

  public void enableAPOP()
  {
    EnableAPOP=true;
  }

  public void disableAPOP()
  {
    EnableAPOP=false;
  }

  public void enableUSERPASS()
  {
    EnableUSERPASS=true;
  }

  public void disableUSERPASS()
  {
    EnableUSERPASS=false;
  }

  public void enableLog(PrintStream LogStream)
  {
    log = LogStream;
  }

  public void setMailFactory(String Name)
  {
    MailFactoryName = Name;
  }

  public void setProperties(Properties p)
  {
    MailFactoryProperties = p;
  }
 
  private MailFactory newMailFactory()
  {
    Class c;
    Class Interfaces[];
    MailFactory m;
    boolean IsMailFactory = false;
    int i;

    if (MailFactoryName == null)
    {
      System.err.println("No mail factory class specified.");
      return null;
    }

    try
    {
      c = Class.forName(MailFactoryName);
    }
    catch (ClassNotFoundException e)
    {
      System.err.println("Class "+MailFactoryName+" not found");
      return null;
    }

    Interfaces = c.getInterfaces();
  
    for (i=0; i<Interfaces.length; i++)
    {
      if (Interfaces[i].getName().equals("rossi.Pop.MailFactory"))
        IsMailFactory = true;
    }

    if (IsMailFactory == false)
    {
      System.err.println("Class "+MailFactoryName+" is not a MailFactory");
      return null;
    }

    try
    {
      m = (MailFactory) c.newInstance();
    }
    catch (InstantiationException e)
    {
      return null;
    }
    catch (IllegalAccessException e)
    {
      return null;
    }

    return m;
  }

  protected void logLogin(String username)
  {
  }

  private void println(String s)
  {
    pout.print(s+"\n");
    if (log != null)
      log.println(SessionIdentifier+" < "+s);
  }
}
