
package rossi.Pop;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

/* This class extends the PopProtocol class by adding the networking
   functionality to it. */

class ServerInstance extends PopProtocol
{
  protected Socket sock;
  protected PrintStream Log;

  ServerInstance(Socket s, InputStream in, OutputStream out, Configuration Config, String HostName, PrintStream Log)
  {
    super(in, out, Config, HostName);
    sock = s;
    this.Log = Log;

    if (Log != null)
      enableLog(Log);

    setMailFactory(Server.MailFactoryName);

    if (Server.EnableLAST)
      enableLAST();
    else
      disableLAST();

    if (Server.EnableUIDL)
      enableUIDL();
    else
      disableUIDL();

    if (Server.EnableTOP)
      enableTOP();
    else
      disableTOP();

    if (Server.EnableAPOP)
      enableAPOP();
    else
      disableAPOP();

    if (Server.EnableUSERPASS)
      enableUSERPASS();
    else
      disableUSERPASS();

    setProperties(Server.ServerProps);
    
    (new Thread(this)).start();
  }

  public void run()
  {
    super.run();
    synchronized (this)
    {
      try
      {
        if (sock != null)
	  sock.close();
      }
      catch (IOException e)
      {
      }
      sock = null;
    }
  }

  protected void logLogin(String username)
  {
    Log.println("Login for "+username+" from: "+sock.getInetAddress().getHostAddress());
  }
}

public class Server implements Runnable
{
  protected static String ConfigFileName;
  protected static String WrapperConfigFileName;
  protected static boolean UserWrapper = false;
  protected static boolean Debug = true;
  protected static TcpdWrapper Wrapper;
  protected static String HostName = "localhost";
  protected static PrintStream Log = null;
  protected static boolean EnableLAST, EnableUIDL, EnableTOP;
  protected static boolean EnableAPOP, EnableUSERPASS;
  protected static String MailFactoryName;
  protected static String Version = "1.0";
  protected static String RegName, RegKey;
  protected static boolean Registered;
  protected static Properties ServerProps = null;
  protected static long TimeoutTime = 15*60*1000;   /* Set to 15 min. */
  private ServerInstance Instance;

  /* This is the POP timeout timer.  It should be reset each time a
     command is received */
  public void run()
  {
    long CurrentTime;
   
    while (true)
    {
      synchronized(Instance)
      {
	CurrentTime = System.currentTimeMillis();
        try
        {
	  Instance.wait(TimeoutTime);
        }
        catch (InterruptedException e)
        {
        }

        if (Instance.sock == null)
          return;

        if ((System.currentTimeMillis() - CurrentTime) > (TimeoutTime*9/10))
        {
          if (Instance.sock != null)
          {
            try
            {
              Instance.sock.close();
              Instance.sock = null;
            }
            catch (IOException e)
            {
            }
          }
          return;
        }
      }
    }
  }

  protected static File LoadProperties()
  {
    String Fname="";
    String LogFname;
    Properties SysProp;
    File f;
    String s;

    SysProp = System.getProperties();
    s = SysProp.getProperty("java.home")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
    f = new File(s);
    if (f.exists())
    {
      Fname = s;
    }
    else
    {
      s = SysProp.getProperty("user.home")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
      f = new File(s);
      if (f.exists())
      {
	Fname = s;
      }
      else
      {
	s = SysProp.getProperty("user.dir")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
	f = new File(s);
	if (f.exists())
	{
	  Fname = s;
	}
	else
	{
	  System.err.println("Cannot find Properties file (Pop3d.pty)");
	  System.exit(1);
	}
      }
    }
    
    ServerProps = new Properties();
    FileInputStream PropIn;
    try
    {
      System.err.println("Loading properties from "+Fname);
      PropIn = new FileInputStream(Fname);
      ServerProps.load(PropIn);
      PropIn.close();
    }
    catch (IOException e)
    {
      System.err.println("Cannot load properties from "+Fname);
      System.exit(1);
    }

    ConfigFileName = ServerProps.getProperty("rossi.Pop.UserList"); 
    UserWrapper = ServerProps.getProperty("rossi.Pop.UseTcpd").equals("true");
    WrapperConfigFileName = ServerProps.getProperty("rossi.Pop.TcpdConfig");
    LogFname = ServerProps.getProperty("rossi.Pop.LogFile");
    HostName = ServerProps.getProperty("rossi.Pop.HostName");
    EnableLAST = ServerProps.getProperty("rossi.Pop.EnableLAST").equals("true");
    EnableUIDL = ServerProps.getProperty("rossi.Pop.EnableUIDL").equals("true");
    EnableTOP = ServerProps.getProperty("rossi.Pop.EnableTOP").equals("true");
    EnableAPOP = ServerProps.getProperty("rossi.Pop.EnableAPOP").equals("true");
    EnableUSERPASS = ServerProps.getProperty("rossi.Pop.EnableUSERPASS").equals("true");

    RegName = ServerProps.getProperty("rossi.Pop.RegName");
    RegKey = ServerProps.getProperty("rossi.Pop.RegKey");
    CheckRegistration(RegName, RegKey);
   
    MailFactoryName = ServerProps.getProperty("rossi.Pop.MailFactory");

    if (LogFname != null)
    {
      try
      {
	Log = new PrintStream(new LogOutputStream(LogFname), true);
      }
      catch (IOException e)
      {
	System.err.println("Failed to enable log file");
      }
    }
    
    if (UserWrapper)
      Wrapper = new TcpdWrapper(WrapperConfigFileName);
    
    return f;
  }

  protected static void CheckRegistration(String RegName, String RegKey)
  {
    if (RegName == null || RegKey == null)
      return;
    /* Make sure we have the MD5 class files */
    try
    {
//      Class.forName("rossi.Pop.MD5");
//      Class.forName("rossi.Pop.MD5_CTX");

      MD5_CTX context = new MD5_CTX();
      byte digest[] = new byte[16];
      String ClearTxt, DigestStr = "", s;
      long ldigest, key;
      byte[] str;
      int len, i;

      ClearTxt = "rossi.Pop.Server "+Version+RegName;

      len = ClearTxt.length();
      str = new byte[len];

      ClearTxt.getBytes(0,len,str,0);
      MD5.MD5Init(context);
      MD5.MD5Update(context, str, len);
      MD5.MD5Final(digest, context);
      ldigest =  ((long)(digest[0] ^ digest[4] ^ digest[8] ^ digest[12]) & 0xff) << 24;
      ldigest |= ((long)(digest[1] ^ digest[5] ^ digest[9] ^ digest[13]) & 0xff) << 16;
      ldigest |= ((long)(digest[2] ^ digest[6] ^ digest[10] ^ digest[14]) & 0xff) << 8;
      ldigest |= ((long)(digest[3] ^ digest[7] ^ digest[11] ^ digest[15]) & 0xff);
      ldigest &= 0xffffffff;
      
      key = Long.parseLong(RegKey);
 
      if (ldigest == key)
        Registered = true;
    }
//    catch (ClassNotFoundException e)
//    {
//      return;
//    }
    catch (NumberFormatException e)
    {
      return;
    }
  }

  public static void main(String args[])
  {
    Configuration Config;
    ServerSocket SSock;
    Socket ns;
    File f;
    long ModTime;
    Server srv;

    System.err.println("J-POP  (C) 1997 Rossi Engineering, All rights reserverd");
    f = LoadProperties();
    ModTime = f.lastModified();

    if (Registered)
      System.err.println("Registered Version "+Version+" Registered to: "+RegName);
    else
      System.err.println("UNREGISTERED Version "+Version);

    Config = new Configuration(ConfigFileName);

    try
    {
      SSock = new ServerSocket(110, 5); 
    }
    catch (IOException e)
    {
      System.err.println(e.getMessage());
      return;
    }
 
    while (true)
    {
      /* Check the registration and exit if a violation, */
      /* We do this in the loop since users can be added */
      /* once the server is running.                     */

      if (Registered == false && Config.UserList.size() > 5)
      {
        System.err.println("Unregistered version does not support this many");
        System.err.println("users.  Please register.");
        System.exit(1);
      }

      try
      {
        /* Wait for a connection */
        ns = SSock.accept();
         
        /* If the properties file has changed, re-load it */
        if (ModTime != f.lastModified())
        {
	  f = LoadProperties();
	  ModTime = f.lastModified();

	  Config = new Configuration(ConfigFileName);
        }

        if (Debug)
          System.err.println(new Date()+" - Connection from: "+ns.getInetAddress().getHostAddress());

        if (Log != null)
          Log.println(new Date()+" - Connection from: "+ns.getInetAddress().getHostAddress());
 
        /* Check that the client is allowed to talk to us. */
        if (UserWrapper)
        {
          if (Wrapper.CheckAuth(ns.getInetAddress()) == false)
          { 
            if (Debug)
              System.err.println("Refusing connection");
            if (Log != null)
              Log.println("Refusing connection");
            ns.close();
            continue;
          }
        }

        srv = new Server();
        srv.Instance = new ServerInstance(ns, ns.getInputStream(), ns.getOutputStream(), Config, HostName, Log);
        (new Thread(srv)).start();
      }
      catch (IOException e)
      {
      }
    }
  }
}
