
package rossi.Pop;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

/* This class extends the PopProtocol class by adding the networking
   functionality to it. */

class TestInstance extends PopProtocol
{
  TestInstance(InputStream in, OutputStream out, Configuration Config, String HostName, PrintStream Log)
  {
    super(in, out, Config, HostName);
    if (Log != null)
      enableLog(Log);

    setMailFactory(Test.MailFactoryName);

    if (Test.EnableLAST)
      enableLAST();
    else
      disableLAST();

    if (Test.EnableUIDL)
      enableUIDL();
    else
      disableUIDL();

    if (Test.EnableTOP)
      enableTOP();
    else
      disableTOP();

    if (Test.EnableAPOP)
      enableAPOP();
    else
      disableAPOP();

    if (Test.EnableUSERPASS)
      enableUSERPASS();
    else
      disableUSERPASS();

    setProperties(Test.TestProps);
    
    (new Thread(this)).start();
  }

  public void run()
  {
    super.run();
  }
}

public class Test
{
  protected static String ConfigFileName;
  protected static String WrapperConfigFileName;
  protected static boolean UserWrapper = false;
  protected static boolean Debug = true;
  protected static TcpdWrapper Wrapper;
  protected static String HostName = "localhost";
  protected static PrintStream Log = null;
  protected static boolean EnableLAST, EnableUIDL, EnableTOP;
  protected static boolean EnableAPOP, EnableUSERPASS;
  protected static String MailFactoryName;
  protected static String Version = "1.0";
  protected static String RegName, RegKey;
  protected static boolean Registered;
  protected static Properties TestProps = null;

  protected static File LoadProperties()
  {
    String Fname="";
    String LogFname;
    Properties SysProp;
    File f;
    String s;

    SysProp = System.getProperties();
    s = SysProp.getProperty("java.home")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
    f = new File(s);
    if (f.exists())
    {
      Fname = s;
    }
    else
    {
      s = SysProp.getProperty("user.home")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
      f = new File(s);
      if (f.exists())
      {
	Fname = s;
      }
      else
      {
	s = SysProp.getProperty("user.dir")+SysProp.getProperty("file.separator")+"Pop3d.pty"; 
	f = new File(s);
	if (f.exists())
	{
	  Fname = s;
	}
	else
	{
	  System.err.println("Cannot find Properties file (Pop3d.pty)");
	  System.exit(1);
	}
      }
    }
    
    TestProps = new Properties();
    FileInputStream PropIn;
    try
    {
      System.err.println("Loading properties from "+Fname);
      PropIn = new FileInputStream(Fname);
      TestProps.load(PropIn);
      PropIn.close();
    }
    catch (IOException e)
    {
      System.err.println("Cannot load properties from "+Fname);
      System.exit(1);
    }

    ConfigFileName = TestProps.getProperty("rossi.Pop.UserList"); 
    UserWrapper = TestProps.getProperty("rossi.Pop.UseTcpd").equals("true");
    WrapperConfigFileName = TestProps.getProperty("rossi.Pop.TcpdConfig");
    LogFname = TestProps.getProperty("rossi.Pop.LogFile");
    HostName = TestProps.getProperty("rossi.Pop.HostName");
    EnableLAST = TestProps.getProperty("rossi.Pop.EnableLAST").equals("true");
    EnableUIDL = TestProps.getProperty("rossi.Pop.EnableUIDL").equals("true");
    EnableTOP = TestProps.getProperty("rossi.Pop.EnableTOP").equals("true");
    EnableAPOP = TestProps.getProperty("rossi.Pop.EnableAPOP").equals("true");
    EnableUSERPASS = TestProps.getProperty("rossi.Pop.EnableUSERPASS").equals("true");

    RegName = TestProps.getProperty("rossi.Pop.RegName");
    RegKey = TestProps.getProperty("rossi.Pop.RegKey");
    CheckRegistration(RegName, RegKey);
   
    MailFactoryName = TestProps.getProperty("rossi.Pop.MailFactory");

    if (LogFname != null)
    {
      try
      {
	Log = new PrintStream(new LogOutputStream(LogFname), true);
      }
      catch (IOException e)
      {
	System.err.println("Failed to enable log file");
      }
    }
    
    if (UserWrapper)
      Wrapper = new TcpdWrapper(WrapperConfigFileName);
    
    return f;
  }

  protected static void CheckRegistration(String RegName, String RegKey)
  {
    if (RegName == null || RegKey == null)
      return;
    /* Make sure we have the MD5 class files */
    try
    {
//      Class.forName("rossi.Pop.MD5");
//      Class.forName("rossi.Pop.MD5_CTX");

      MD5_CTX context = new MD5_CTX();
      byte digest[] = new byte[16];
      String ClearTxt, DigestStr = "", s;
      long ldigest, key;
      byte[] str;
      int len, i;

      ClearTxt = "rossi.Pop.Server "+Version+RegName;

      len = ClearTxt.length();
      str = new byte[len];

      ClearTxt.getBytes(0,len,str,0);
      MD5.MD5Init(context);
      MD5.MD5Update(context, str, len);
      MD5.MD5Final(digest, context);
      ldigest =  ((long)(digest[0] ^ digest[4] ^ digest[8] ^ digest[12]) & 0xff) << 24;
      ldigest |= ((long)(digest[1] ^ digest[5] ^ digest[9] ^ digest[13]) & 0xff) << 16;
      ldigest |= ((long)(digest[2] ^ digest[6] ^ digest[10] ^ digest[14]) & 0xff) << 8;
      ldigest |= ((long)(digest[3] ^ digest[7] ^ digest[11] ^ digest[15]) & 0xff);
      ldigest &= 0xffffffff;
      
      key = Long.parseLong(RegKey);
 
      if (ldigest == key)
        Registered = true;
    }
//    catch (ClassNotFoundException e)
//    {
//      return;
//    }
    catch (NumberFormatException e)
    {
      return;
    }
  }

  public static void main(String args[])
  {
    Configuration Config;
    Socket ns;
    File f;
    long ModTime;

    System.err.println("J-POP  (C) 1997 Rossi Engineering, All rights reserverd");
    f = LoadProperties();
    ModTime = f.lastModified();

    if (Registered)
      System.err.println("Registered Version "+Version+" Registered to: "+RegName);
    else
      System.err.println("UNREGISTERED Version "+Version);

    Config = new Configuration(ConfigFileName);

 
    new TestInstance(System.in, System.out, Config, HostName, Log);
  }
}
