
package rossi.Pop;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

class UnixMailLockFile
{
  File lockfile = null;
  int Timeout;
  byte[] Keyvalue;

  public UnixMailLockFile()
  {
    Keyvalue = new byte[4];
    Timeout = 20;
  }

  public void setKey(int Key)
  {
    Keyvalue[0] = (byte)((Key >>> 24) & 0xff);
    Keyvalue[1] = (byte)((Key >>> 16) & 0xff);
    Keyvalue[2] = (byte)((Key >>> 8) & 0xff);
    Keyvalue[3] = (byte)(Key & 0xff);
  }

  public void setFile(String Fname)
  {
    lockfile = new File(Fname);
  }

  public void setTimeout(int t)
  {
    Timeout = t;
  }

  public boolean getLock()
  {
    long existtime; 
    long starttime;
    long InitialTimeStamp;
    FileOutputStream out;
    FileInputStream in;
    int ExceptionCount = 0;
  
    InitialTimeStamp = lockfile.lastModified();
    starttime = System.currentTimeMillis();

    while (true)
    {
      existtime = System.currentTimeMillis();
      if (lockfile.exists())
      {
        try
        {
          Thread.sleep(1000);
        }
        catch (InterruptedException e)
        {
        }
 
        if ((existtime - starttime) < 1000 * Timeout)
        {
          continue; 
        }
        
        if (InitialTimeStamp != lockfile.lastModified())
        {
          InitialTimeStamp = lockfile.lastModified();
          starttime = System.currentTimeMillis();
          continue; 
        }

        lockfile.delete();
        InitialTimeStamp = lockfile.lastModified();
        starttime = System.currentTimeMillis();
        continue;
      }

      /* Lock file didn't exist */

      try
      {
        long timestamp;
        byte[] testdata = new byte[4];

        out = new FileOutputStream(lockfile);
        out.write(Keyvalue);
        out.close();
        timestamp = lockfile.lastModified();
 

        if ((System.currentTimeMillis() - existtime) > 250)
        {
          continue; 
        }

        try
        {
          Thread.sleep(1000);
        }
        catch (InterruptedException e)
        {
        }

        if (lockfile.length() != 4 || lockfile.lastModified() != timestamp)
        {
          continue;
        }

        in = new FileInputStream(lockfile);
        in.read(testdata); 
        in.close();

        if (testdata[0] != Keyvalue[0] ||
            testdata[1] != Keyvalue[1] ||
            testdata[2] != Keyvalue[2] ||
            testdata[3] != Keyvalue[3])
        {
          continue;
        }
 
      }
      catch (IOException e)
      {
        ExceptionCount++;

        if (ExceptionCount > 10)
          return false;

        continue; 
      }
      return true; 
    }
  }

  public void releaseLock()
  {
    lockfile.delete();
  }
}


public class UnixMailFactory extends InputStream implements MailFactory
{
  /* Stuff for MailFactory */
  UnixMailSpec MailSpecs[];
  Properties OurProps, DefaultProperties;
  String[] PropNames = {"rossi.Pop.UnixMailFactory.TempDir",
                        "rossi.Pop.UnixMailFactory.LockFileExt",
                        "rossi.Pop.UnixMailFactory.UseIndex"};

  /* Stuff for emulating InputStream */
  int MessageNumber;
  int MessageBytesRead;
  RandomAccessFile MailInput;
  String MailFileName;
  String TmpMailFileName;
  String IndexName;
  String TmpDirName;
  UnixMailLockFile Lockf;
  boolean UseIndex;

  public UnixMailFactory()
  {
    DefaultProperties = new Properties();
    DefaultProperties.put("rossi.Pop.UnixMailFactory.TempDir",File.separator+"tmp");
    DefaultProperties.put("rossi.Pop.UnixMailFactory.LockFileExt","lock");
    DefaultProperties.put("rossi.Pop.UnixMailFactory.UseIndex","false");

    OurProps = DefaultProperties;
  }

  public boolean setMailbox(String Path)
  {
    RandomAccessFile MailFile;
    int StartOfLine = 0;
    byte[] Line;
    int LineLen = 0;
    boolean MsgDelim = false;
    Vector v;
    UnixMailSpec us = null;
    int NumMsgs = 0;
    int LineCounter = 0;
    int ByteCounter = 0;
    int i;
    MD5_CTX context = null;
    byte digest[];
    byte buffer[];
    int TotalBytesRead = 0, BytesRead = 0, StartOffset = 0;
    boolean OnHeaders = false;
    boolean EndOfData = false;
    boolean StatusLine = false;

    v = new Vector();
    digest = new byte[16];
    buffer = new byte[65536];
    Line = new byte[256];
    MailFileName = Path;
    Lockf = new UnixMailLockFile();
    Lockf.setFile(Path+"."+OurProps.getProperty("rossi.Pop.UnixMailFactory.LockFileExt","lock"));
    Lockf.setKey(147165);
    Lockf.setTimeout(60);

    TmpDirName = OurProps.getProperty("rossi.Pop.UnixMailFactory.TempDir",File.separator+"tmp");
    UseIndex = OurProps.getProperty("rossi.Pop.UnixMailFactory.UseIndex",File.separator+"false").equals("true");

    /* Obtain a lock */

    if (Lockf.getLock() == false)
      return false;

    /* Create a temp. copy of the mailbox */
    SetTempFileName();

    /* Special case, if mailbox doesn't exist */
    if (new File(MailFileName).exists() == false)
    {
      try
      {
        FileOutputStream out = new FileOutputStream(TmpMailFileName);
        out.close();
      }
      catch (IOException e)
      {
	Lockf.releaseLock();  /* Release the lock */
	return false;
      }
    }
    else  /* Normal case, get a copy of the mailbox */
    {
      if (CopyFile(MailFileName, 0, TmpMailFileName) == false)
      {
	Lockf.releaseLock();  /* Release the lock */
	return false;
      }
    }

    Lockf.releaseLock();  /* Release the lock */


    /* Open a FileInputStream for the mailbox */
   
    try
    {
      MailFile = new RandomAccessFile(TmpMailFileName, "r");
      
      if (UseIndex)
      {
        readIndex(v, MailFile);
        TotalBytesRead = (int) MailFile.getFilePointer();
      }

      LineLen = 0;
      while (true)
      {
        /* Read in a line the hard way.  Because it is faster than
           calling readLine() -- way faster.  Apparently readLine()
           reads files one byte at a time.  This is very inefficent.
           Also, when we read the data, don't make String objects
           out of it since this is a very time consuming operation
           as well.  Do the equivienlt of our string compares directly
           on the byte array.  Much faster.  
 
           Come to think of it, readLine() returns a String object.
           That could also be part of its slowness.  These techniques
           used here result in over an 100 times speed improvement. */

        if (StartOffset == BytesRead)
        {
          StartOffset = 0;
          BytesRead = MailFile.read(buffer);
          if (BytesRead <= 0)
          {
            StartOfLine = TotalBytesRead;  /* Point to end of file */
            EndOfData = true;
          }
          else
          {
            TotalBytesRead += BytesRead;
            continue;
          }
        }
	else
	{
	  for (i=StartOffset;i<BytesRead;i++)
          {
	    if (buffer[i] == '\n')
	      break;
            Line[LineLen++] = buffer[i];

            /* If line is longer than 200 characters, then break it up
               for our purposes, try not to split on an 'F' that 
               could make later code think it was the start of a new
               message, but if we can't do it by 256 characters, split
               it anyways. */

            if ((LineLen > 200 && i < (buffer.length + 1) 
                && buffer[i+1] != 'F') || LineLen == 256)
            {
              /* Since this isn't a  ** real ** line, subtract off the size
                 of the EOL characters */
              ByteCounter -= 2;
              break;
            }
          }

	  if (i == BytesRead)
	  {
	    StartOffset = i;
	    continue;
	  }
	  else  /* Found newline */
	  {
            if (i > 0 && buffer[i-1] == '\r')
            {
	      LineLen = LineLen - 1;
              StartOfLine = TotalBytesRead - BytesRead + i - LineLen - 1;
            }
            else
            {
              StartOfLine = TotalBytesRead - BytesRead + i - LineLen;
            }
	    StartOffset = i+1;
	  }
	}
        /* Done reading in a line */

        if (LineLen > 5 && Line[0] == 'F' && Line[1] == 'r' 
                        && Line[2] == 'o' && Line[3] == 'm'
                        && Line[4] == ' ')
          MsgDelim = true;
        else
          MsgDelim = false;

        if (MsgDelim || EndOfData)
        {
          /* Finish up processing of the last message */
          if (v.size() != 0 && us != null)
          {
            String s;

            us.Size = ByteCounter;
            us.EndOffset = StartOfLine;
            MD5.MD5Final(digest, context);
            us.UniqID = "";         
	    for (i = 0; i < 16; i++)
	    {
	      s = Integer.toHexString((int)digest[i]&0x000000ff);
	      if (s.length() == 1)
	        s = "0"+s;
	      us.UniqID = us.UniqID+s;
	    }
          }

          if (EndOfData == false)
          {
            /* Start processing the next one */
            context = new MD5_CTX();
            MD5.MD5Init(context);
            LineCounter = 0;
            ByteCounter = 0;
            us = new UnixMailSpec(); 
            us.StartOffset = StartOfLine;
            us.Delete=false;
            us.Read=false;
            us.ReadPreviously = false;
            us.StatusStartOffset = 0;
            us.StatusEndOffset = 0;
            v.addElement(us);
            OnHeaders = true;
          }
          else
            break;
        }
        LineCounter++;
        ByteCounter += (LineLen + 2);

        if (LineLen == 0)
        {
          OnHeaders = false;
          if (us.StatusEndOffset == 0)
          {
            us.StatusStartOffset = StartOfLine;
            us.StatusEndOffset = StartOfLine;
          }
        }

        StatusLine = false;
        if (OnHeaders && LineLen > 8 && Line[0] == 'S' && Line[1] == 't' 
                        && Line[2] == 'a' && Line[3] == 't'
                        && Line[4] == 'u' && Line[5] == 's'
                        && Line[6] == ':' && Line[7] == ' ')
        {
          StatusLine = true;
          if (Line[8] == 'R')
          {
            us.ReadPreviously = true;
            us.Read = true;
            us.StatusStartOffset = StartOfLine;
            us.StatusEndOffset = StartOfLine + LineLen;
          }
        }

        if (context != null && OnHeaders && !StatusLine)
        {
          MD5.MD5Update(context, Line, LineLen);
        }
        LineLen = 0;
      }
 
      MailSpecs = new UnixMailSpec[v.size()];

      System.err.println("Total of "+v.size()+" messages");
      for (i=0; i<v.size(); i++)
      {
        MailSpecs[i] = (UnixMailSpec) (v.elementAt(i));
      }

      MailFile.close();
    }
    catch (IOException e)
    {
      return false;
    }
    return true;
  }

  private void readIndex(Vector v, RandomAccessFile in)
  {
    File IndexFile;
    UnixMailSpec us;
    FileInputStream IndexIn;
    DataInputStream IndexDataIn;
    byte[] buffer = new byte[32];
    int MailboxSize = 0;
    int GoodIndicies;
    int i = 0;
 
    IndexFile = new File(IndexName);
    if (IndexFile.exists() == false)
      return;


    try
    {
      IndexIn = new FileInputStream(IndexFile);
      IndexDataIn = new DataInputStream(IndexIn);

      MailboxSize = (int)in.length();

      while (true)
      {
        us = new UnixMailSpec();
        us.StartOffset = IndexDataIn.readInt();
        us.StatusStartOffset = IndexDataIn.readInt();
        us.StatusEndOffset = IndexDataIn.readInt();
        us.EndOffset = IndexDataIn.readInt();
	us.Size = IndexDataIn.readLong();
        IndexDataIn.readFully(buffer);
        us.UniqID = new String(buffer, 0, 0, buffer.length);
	us.Delete = IndexDataIn.readBoolean();
	us.ReadPreviously = IndexDataIn.readBoolean();
	us.Read = IndexDataIn.readBoolean();
        v.addElement(us);
      }
    }
    catch (EOFException e)
    {
    }
    catch (IOException e)
    {
    }


    try
    {
      for (i=0; i<v.size(); i++)
      {
        us = (UnixMailSpec) v.elementAt(i);

        if (us.EndOffset > MailboxSize)
          break;

        in.seek(us.StartOffset);
        in.readFully(buffer, 0, 5);
        if ((new String(buffer, 0, 0, 5)).equals("From ") == false)
          break;
      }
    }
    catch (IOException e)
    {
    }


    try
    {
      if (i > 0)
      {
        while (true)
          v.removeElementAt(i-1);
      }
      else
        v.removeAllElements();
    }
    catch (ArrayIndexOutOfBoundsException e)
    {
    }
 

    try
    {
      if (i > 1)
      {
	us = (UnixMailSpec) v.elementAt(i-2);
	in.seek(us.EndOffset);
      }
      else
      {
	in.seek(0); 
      }
    }
    catch (IOException e)
    {
    }
  }
  
  private void SetTempFileName()
  {
    String Name;
    File f;
    int i;
    File TmpDir;
    File TmpMailFile;

    Name = new File(MailFileName).getName();
    TmpDir = new File(TmpDirName);

    IndexName = (new File(TmpDir, Name+".ndx")).getAbsolutePath();

    for (i=0; i<10; i++)
    {
      TmpMailFile = new File(TmpDir, Name+"."+i);
      TmpMailFileName = TmpMailFile.getAbsolutePath();
      f = new File(TmpMailFileName);
      if (f.exists() == false)
        break;
    } 
  }

  private boolean CopyFile(String src, int StartOffset, String dest)
  {
    FileInputStream in;
    FileOutputStream out;
    int BytesRead;
    byte[] buffer = new byte[1024];

    try
    {
      in = new FileInputStream(src);
      out = new FileOutputStream(dest);
      in.skip(StartOffset);
      while(true)
      {
        BytesRead = in.read(buffer);
        if (BytesRead == -1)
          break;
        out.write(buffer, 0, BytesRead);
      }
      in.close();
      out.close();
    }
    catch (IOException e)
    { 
      return false;
    }
    return true;
  }

  /* Returns a vector containing a bunch of MailSpec objects */
  public MailSpec[] getMailList()
  {
    return MailSpecs;
  }

  /* Returns an InputStream for reading actual message data */
  public InputStream getMessageInputStream(int MsgNum)
  {
    MessageNumber = MsgNum;
    MessageBytesRead = 0;
    try 
    {
      MailInput = new RandomAccessFile(MailFileName, "r");
      MailInput.seek(MailSpecs[MessageNumber].StartOffset);
    }
    catch (IOException e)
    {
      return null;
    }

    return this;
  }

 
  public void updateMaildrop()   /* Update the mailbox */
  {
    String RecentMailFileName;
    RandomAccessFile in = null;
    FileOutputStream out;
    DataOutputStream dout;
    byte[] buffer = new byte[1024];
    int BytesRead;
    int BytesToRead;
    int BlockSize;
    int MsgNo, i;
    String ReadMarker = "Status: RO";
    UnixMailSpec IndexMailSpecs[] = null;
    int IndexBytesWritten = 0;
    int IndexMsgNo = 0;

    /* Special case -- empty mailbox */
    if (MailSpecs.length == 0)
    {
      new File(TmpMailFileName).delete();
      return;
    }

    /* First, obtain a lock */ 
    if (Lockf.getLock() == false)
      return;

    if (UseIndex)
      IndexMailSpecs = new UnixMailSpec[MailSpecs.length];

    /* Did the mailfile shrink ? If so we'd better not update */
 
    if ((new File(MailFileName).length()) < MailSpecs[MailSpecs.length - 1].EndOffset)
    {
      System.err.println("Mailbox shrank unexpectedly");
      new File(TmpMailFileName).delete();
      /* Release the lock */
      Lockf.releaseLock();
      return;
    }

    /* Copy anything that came in recently to a Temp File */
 
    RecentMailFileName = TmpMailFileName+"t";
    CopyFile(MailFileName, 
             MailSpecs[MailSpecs.length - 1].EndOffset, 
             RecentMailFileName);

    /* Check out the recent mail.  If the file doesn't start with a
       valid message separator, then there is a problem and we should
       abort the update. */

    
    try
    {
      in = new RandomAccessFile(RecentMailFileName, "r");
      if (in.length() > 0)
      {
	in.readFully(buffer, 0, 5);
	if ((new String(buffer, 0, 0, 5)).equals("From ") == false)
	{
	  System.err.println("Mailbox unexpectedly modified");
	  in.close();
	  /* Delete the temp files */
	  new File(RecentMailFileName).delete();
	  new File(TmpMailFileName).delete();
	  /* Release the lock */
	  Lockf.releaseLock();
	  return;
	}
      }
      /* Bug, wasn't closing file, this would cause later delete to fail */
      in.close();  /* Close in any case */
    }
    catch (IOException e)
    {
       /* Something smells funny here, so just give up. */
       return;
    }

    /* Ok, now that we have that, proceed with the update */
 
    try
    {
      in = new RandomAccessFile(TmpMailFileName, "r");
      out = new FileOutputStream(MailFileName);

      for (MsgNo = 0; MsgNo < MailSpecs.length; MsgNo++)
      {
         if (MailSpecs[MsgNo].Delete == false)
         {
           if (UseIndex)
	   {
	     IndexMailSpecs[IndexMsgNo] = new UnixMailSpec();
	     IndexMailSpecs[IndexMsgNo].StartOffset = IndexBytesWritten;
	     IndexMailSpecs[IndexMsgNo].StatusStartOffset = MailSpecs[MsgNo].StatusStartOffset 
							  - MailSpecs[MsgNo].StartOffset 
							  + IndexBytesWritten; 

	     if (MailSpecs[MsgNo].Read)
	     {
	       IndexMailSpecs[IndexMsgNo].StatusEndOffset = IndexMailSpecs[IndexMsgNo].StatusStartOffset + ReadMarker.length();
	     }  
	     else
	     {
	       IndexMailSpecs[IndexMsgNo].StatusEndOffset = IndexMailSpecs[IndexMsgNo].StatusStartOffset 
							  + MailSpecs[MsgNo].StatusEndOffset 
							  - MailSpecs[MsgNo].StatusStartOffset;
   
	     }
	     IndexMailSpecs[IndexMsgNo].EndOffset = IndexMailSpecs[IndexMsgNo].StatusEndOffset 
						  + MailSpecs[MsgNo].EndOffset 
						  - MailSpecs[MsgNo].StatusEndOffset;          
	     IndexMailSpecs[IndexMsgNo].Size = MailSpecs[MsgNo].Size; 
	     IndexMailSpecs[IndexMsgNo].UniqID = MailSpecs[MsgNo].UniqID;
	     IndexMailSpecs[IndexMsgNo].Delete = MailSpecs[MsgNo].Delete;
	     IndexMailSpecs[IndexMsgNo].ReadPreviously = MailSpecs[MsgNo].ReadPreviously;
	     IndexMailSpecs[IndexMsgNo].Read = MailSpecs[MsgNo].Read;
	     IndexBytesWritten += IndexMailSpecs[IndexMsgNo].EndOffset - IndexMailSpecs[IndexMsgNo].StartOffset;
	     IndexMsgNo++;
	   }

           in.seek(MailSpecs[MsgNo].StartOffset);

           /* Write the message headers */

           BlockSize = MailSpecs[MsgNo].StatusStartOffset - 
                       MailSpecs[MsgNo].StartOffset;

           while (BlockSize != 0)
           {
             BytesToRead = (BlockSize > buffer.length) ? buffer.length : BlockSize;
             BytesRead = in.read(buffer, 0, BytesToRead);
             if (BytesRead == -1)
               break;
             BlockSize -= BytesRead;
             out.write(buffer, 0, BytesRead);
           }
            
           if (MailSpecs[MsgNo].Read)
           {
             /* Write the status header */

             ReadMarker.getBytes(0, ReadMarker.length(), buffer, 0);
             out.write(buffer, 0, ReadMarker.length());
             in.seek(MailSpecs[MsgNo].StatusEndOffset);
           }
           else
           {
             /* Write the status header */

	     BlockSize = MailSpecs[MsgNo].StatusEndOffset - 
			 MailSpecs[MsgNo].StatusStartOffset;

	     while (BlockSize != 0)
	     {
	       BytesToRead = (BlockSize > buffer.length) ? buffer.length : BlockSize;
	       BytesRead = in.read(buffer, 0, BytesToRead);
	       if (BytesRead == -1)
		 break;
	       BlockSize -= BytesRead;
	       out.write(buffer, 0, BytesRead);
	     }
           }

           /* Write the message body */

           BlockSize = MailSpecs[MsgNo].EndOffset - 
                       MailSpecs[MsgNo].StatusEndOffset;

           while (BlockSize != 0)
           {
             BytesToRead = (BlockSize > buffer.length) ? buffer.length : BlockSize;
             BytesRead = in.read(buffer, 0, BytesToRead);
             if (BytesRead == -1)
               break;
             BlockSize -= BytesRead;
             out.write(buffer, 0, BytesRead);
           }
        }
      }

      in.close();  /* Close the TmpMailFileName file */

      in = new RandomAccessFile(RecentMailFileName, "r");

      BlockSize = (int) new File(RecentMailFileName).length();
 
      while (BlockSize != 0)
      {
         BytesToRead = (BlockSize > buffer.length) ? buffer.length : BlockSize;
	 BytesRead = in.read(buffer, 0, BytesToRead);
	 if (BytesRead == -1)
	   break;
	 BlockSize -= BytesRead;
	 out.write(buffer, 0, BytesRead);
      }
      in.close();
      out.close();
      
      /* Delete the temp files */
      new File(RecentMailFileName).delete();
      new File(TmpMailFileName).delete();


      /* Write out the index file */
      
      if (UseIndex) 
      {
	out = new FileOutputStream(IndexName);
	dout = new DataOutputStream(out);

	for (i=0; i<IndexMsgNo; i++)
	{
	  dout.writeInt(IndexMailSpecs[i].StartOffset);
	  dout.writeInt(IndexMailSpecs[i].StatusStartOffset);
	  dout.writeInt(IndexMailSpecs[i].StatusEndOffset);
	  dout.writeInt(IndexMailSpecs[i].EndOffset);
	  dout.writeLong(IndexMailSpecs[i].Size);
	  dout.writeBytes(IndexMailSpecs[i].UniqID);
	  dout.writeBoolean(IndexMailSpecs[i].Delete);
	  dout.writeBoolean(IndexMailSpecs[i].ReadPreviously);
	  dout.writeBoolean(IndexMailSpecs[i].Read);
	}
	dout.flush();
	out.close();
      }
    }
    catch (IOException e)
    {
      System.err.println("IOException writing mailbox: "+e.getMessage());
    }

    /* Release the lock */
    Lockf.releaseLock();
  } 

  public void setProperties(Properties p)
  {
    if (p != null)
      OurProps = p;
  }

  public String[] listPropertyNames()
  {
    return PropNames; 
  }

  public String checkProperties()
  {
    File f;
    String s;

    if (OurProps == null)
      return "Properties not set";
    
    s = OurProps.getProperty("rossi.Pop.UnixMailFactory.TempDir");
    if (s == null || s.equals(""))
      return "Must specify TempDir";

    f = new File(s);
    if (f.exists() == false)
      return "TempDir does not exist";

    if (f.isDirectory() == false)
      return "TempDir is not a directory";
 
    s = OurProps.getProperty("rossi.Pop.UnixMailFactory.UseIndex");
    if (s.equals("true") == false && s.equals("false") == false)
      return "UseIndex must be true or false";

    return null;
  }

  public Properties getDefaultProperties()
  {
    return DefaultProperties;
  }

  /* Stuff for InputStream */
  public int available()
  {
    return MailSpecs[MessageNumber].EndOffset 
          - MailSpecs[MessageNumber].StartOffset - MessageBytesRead;
  }

  public void close() throws IOException
  {
    MailInput.close();
  }

  public synchronized void mark(int i)
  {
  }

  public synchronized void reset()
  {
  }

  public boolean marksSupported()
  {
    return false;
  }

  public long skip(long l) throws IOException
  {
    if ((MessageBytesRead + (int) l) <= 
        MailSpecs[MessageNumber].EndOffset - MailSpecs[MessageNumber].StartOffset)
    {
      MailInput.seek(MailSpecs[MessageNumber].StartOffset+MessageBytesRead+(int) l);
      MessageBytesRead += (int) l;
      return l;
    }
    else
      throw new IOException();
  }

  public int read() throws IOException
  {
    int l;

    if (available() == 0)
      return -1;
    
    l = MailInput.read();
    if (l >= 0)
      MessageBytesRead++;

    return l;
  }

  public int read(byte[] buffer) throws IOException
  {
    int ReqLen = buffer.length;
    int Avail = available();
    int l;

    if (Avail == 0)
      return -1;
 
    if (ReqLen > Avail)
      ReqLen = Avail;

    l = MailInput.read(buffer, 0, ReqLen);
    if (l > 0)
      MessageBytesRead+=l;
    if (l == 0)
      l = -1;
    return l;
  }

  public int read(byte[] buffer, int off, int len) throws IOException
  {
    int ReqLen = len;
    int Avail = available();
    int l;

    if (Avail == 0)
      return -1;
 
    if (ReqLen > Avail)
      ReqLen = Avail;

    l = MailInput.read(buffer, off, ReqLen);
    if (l > 0)
      MessageBytesRead+=l;
    if (l == 0)
      l = -1;
    return l;
  }

  /* Make sure we clean up a bit */

  /* Commented out because it was causing problems. */
//  protected void finalize() throws IOException
//  {
//    Lockf.releaseLock();  /* Release the lock */
//    close();
//    new File(TmpMailFileName).delete();
//  }
}
